from sqlalchemy import Column, Integer, DateTime
from sqlalchemy.ext.declarative import declarative_base

from openmodule.database.database import run_env_py

Base = declarative_base()


class DatabaseInvalidTimezoneTestModel(Base):
    __tablename__ = "test_timezone_invalid"
    id = Column(Integer, primary_key=True, autoincrement=True)
    datetime = Column(DateTime, nullable=True)


run_env_py([Base])



import uuid

from sqlalchemy import Column, String, Integer, Date, Time
from sqlalchemy.ext.declarative import declarative_base

from openmodule.database.custom_types import TZDateTime

Base = declarative_base()


class DatabaseTestModel(Base):
    __tablename__ = "test"
    id = Column(String, default=lambda: str(uuid.uuid4()), unique=True, primary_key=True)
    value1 = Column(Integer, default=1)
    value2 = Column(Integer, default=2)
    string = Column(String, default="initial")


class DatabaseTimezoneTestModel(Base):
    __tablename__ = "test_timezone"
    id = Column(Integer, primary_key=True, autoincrement=True)
    tz_datetime = Column(TZDateTime, nullable=True)
    date = Column(Date, nullable=True)
    time = Column(Time, nullable=True)



from datetime import datetime

from pydantic.main import BaseModel
from sqlalchemy import DateTime

from openmodule import config


class CheckingOpenModuleModel(type(BaseModel), type):
    def __new__(mcs, name, bases, dct):
        cls = super().__new__(mcs, name, bases, dct)

        # noinspection PyUnresolvedReferences
        for field_name, field in cls.__fields__.items():
            if field.type_ == datetime:
                assert "_timezone_validator" in field.class_validators, (
                    "datetime fields must use the timezone_validator to ensure all datetime values "
                    "are always naive. Otherwise runtime errors may occur depending on the isoformat used.\n"
                    f"In class {name}, please add:\n"
                    f'  _tz_{field_name} = timezone_validator("{field_name}")'
                )

        return cls


def check_invalid_database_column_type(typ):
    if config.run_checks():
        assert not isinstance(typ, DateTime), "Do NOT use DateTime fields, use TZDateTime fields instead"



import logging
import os
import socket
import warnings
from types import SimpleNamespace
from typing import Optional, Any
from unittest import TestCase

from pydantic import ValidationError

from openmodule import config
from openmodule.config import validate_config_module, LazySettings, override_settings, settings, \
    override_context
from openmodule.models.base import OpenModuleModel
from openmodule_test.files import temp_file
from openmodule_test.zeromq import ZMQTestMixin
from tests.resources.configs import test_config


class EnvTestCase(TestCase):
    initial_env: dict

    def env(self, value, key="t"):
        os.environ[key] = value

    def setUp(self) -> None:
        self.initial_env = os.environ.copy()
        super().setUp()

    def tearDown(self):
        os.environ.clear()
        os.environ.update(self.initial_env)
        super().tearDown()


class ConfigTest(EnvTestCase):
    def test_int(self):
        self.env("1")
        self.assertEqual(1, config.int("t", 0))

        self.env("a")
        self.assertEqual(0, config.int("t", 0))

        self.env("0x16")
        self.assertEqual(0, config.int("t", 0))

        with self.assertRaises(AssertionError):
            config.int("t", 1.1)

    def test_float(self):
        self.env("1")
        self.assertEqual(1, config.float("t", 0))
        self.env("1.0")
        self.assertEqual(1, config.float("t", 0))
        self.env("1.1")
        self.assertEqual(1.1, config.float("t", 0))

        self.env("a")
        self.assertEqual(0, config.float("t", 0))

        self.env("0x16")
        self.assertEqual(0, config.float("t", 0))

        with self.assertRaises(AssertionError):
            config.float("t", "a")

    def test_str(self):
        self.env("1")
        self.assertEqual("1", config.string("t", "a"))

        self.env(" \"'1' \r\n\t")  # config.string strips unnecessary whitespace and quotes
        self.assertEqual("1", config.string("t", "a"))

        with self.assertRaises(AssertionError):
            config.string("t", 1.0)  # noqa

    def test_bool(self):
        self.env("1")
        self.assertTrue(config.bool("t", False))
        self.env("y")
        self.assertTrue(config.bool("t", False))
        self.env("Y")
        self.assertTrue(config.bool("t", False))
        self.env("t")
        self.assertTrue(config.bool("t", False))
        self.env("true")
        self.assertTrue(config.bool("t", False))
        self.env("yes")
        self.assertTrue(config.bool("t", False))

        # empty string is default value
        self.env("")
        self.assertTrue(config.bool("t", True))
        self.assertFalse(config.bool("t", False))

        with self.assertRaises(AssertionError):
            config.bool("t", "a")  # noqa

    def test_log_level(self):
        self.env("", "LOG_LEVEL")
        self.assertEqual(logging.INFO, config.log_level())  # info is the default value
        self.env("info", "LOG_LEVEL")
        self.assertEqual(logging.INFO, config.log_level())
        self.env("WARN", "LOG_LEVEL")
        self.assertEqual(logging.WARNING, config.log_level())
        self.env("ERROR", "LOG_LEVEL")
        self.assertEqual(logging.ERROR, config.log_level())
        self.env("CRITICAL", "LOG_LEVEL")
        self.assertEqual(logging.CRITICAL, config.log_level())
        self.env("VERBOSE", "LOG_LEVEL")
        self.assertEqual(logging.DEBUG, config.log_level())
        self.env("DEbUG", "LOG_LEVEL")
        self.assertEqual(logging.DEBUG, config.log_level())

    def test_version(self):
        # read version from env first
        self.env("1.0.0", "VERSION")
        self.assertEqual("1.0.0", config.version())

        # if it is not present, read it from VERSION file
        self.env("", "VERSION")
        try:
            with open("VERSION", "w") as f:
                f.write("1.0.0\n")
            self.assertEqual("1.0.0", config.version())
        finally:
            os.unlink("VERSION")

        # if we find nothing, version === "unknown"
        self.assertEqual("unknown", config.version())

    def test_resource(self):
        # no auth credentials set -> use hostname which is nice during development
        self.assertEqual(socket.gethostname(), config.resource())

        self.env("TEST123", "AUTH_RESOURCE")
        self.assertEqual("TEST123", config.resource())


class ConfigYamlTest(EnvTestCase):
    class ExampleModel(OpenModuleModel):
        string: str = "default"

    class NonConstructable(OpenModuleModel):
        string: str

    def test_path_from_env(self):
        with temp_file("string: not-default") as path:
            self.env(path, "CONFIG_YAML")
            instance = config.yaml(self.ExampleModel)
            self.assertEqual('not-default', instance.string)

    def test_default_model(self):
        instance = config.yaml(self.ExampleModel)
        self.assertEqual("default", instance.string)

    def test_invalid_default_model(self):
        # we do not protect against any validation errors, but simply log them
        # in any case this is mostly a developer's fault, and hardly ever a user fault
        with self.assertRaises(ValidationError):
            with self.assertLogs() as cm:
                config.yaml(self.NonConstructable)

        self.assertIn("wrong with the configuration", str(cm.output))

    def test_yaml_ignores_custom_constructors(self):
        yaml = """\
        test: string
        some: !<CustomObject>
            a: b
        """

        class TestModel(OpenModuleModel):
            test: str
            some: Optional[Any]

        with temp_file(yaml) as f:
            model = config.yaml(TestModel, f)

        self.assertIsNone(model.some, msg="Not parsed custom elements must be none")
        self.assertEqual(model.test, "string")

    def test_empty_yml(self):
        yaml = ""

        class TestModel(OpenModuleModel):
            some: Optional[Any]

        with temp_file(yaml) as f:
            model = config.yaml(TestModel, f)

        self.assertIsNone(model.some, msg="Not parsed custom elements must be none")



class LegacyConfigSupportTest(EnvTestCase):
    def test_broker_pub(self):
        # nothing found -> defaults
        self.assertEqual("tcp://127.0.0.1:10200", config.broker_pub())

        # legacy configs
        self.env("1.2.3.4", "BROKER_HOST")
        self.env("1234", "BROKER_PUB_PORT")
        self.assertEqual("tcp://1.2.3.4:1234", config.broker_pub())

        # new configs overrule
        self.env("inproc://test-pub", "BROKER_PUB")
        self.env("1.2.3.4", "BROKER_HOST")
        self.env("1234", "BROKER_PUB_PORT")
        self.assertEqual("inproc://test-pub", config.broker_pub())

    def test_broker_sub(self):
        # nothing found -> defaults
        self.assertEqual("tcp://127.0.0.1:10100", config.broker_sub())

        # legacy configs
        self.env("1.2.3.4", "BROKER_HOST")
        self.env("1234", "BROKER_SUB_PORT")
        self.assertEqual("tcp://1.2.3.4:1234", config.broker_sub())

        # new configs overrule
        self.env("inproc://test-pub", "BROKER_SUB")
        self.env("1.2.3.4", "BROKER_HOST")
        self.env("1234", "BROKER_SUB_PORT")
        self.assertEqual("inproc://test-pub", config.broker_sub())

    def test_missing_keys_in_config_cause_assertion(self):
        fake_config = SimpleNamespace()
        fake_config.NAME = "test"
        with self.assertRaises(AssertionError):
            validate_config_module(fake_config)

    def test_legacy_broker_configs_cause_warning(self):
        with warnings.catch_warnings(record=True) as w:
            warnings.simplefilter("always")

            fake_config = SimpleNamespace()
            fake_config.NAME = "test"
            fake_config.RESOURCE = "test"
            fake_config.VERSION = "test"
            fake_config.DEBUG = False
            fake_config.TESTING = True
            fake_config.BROKER_PUB_PORT = "10000"
            validate_config_module(fake_config)

        self.assertEqual(1, len(w))
        self.assertTrue(issubclass(w[0].category, DeprecationWarning))
        self.assertIn("BROKER_PUB", str(w[0].message))


class LazyConfigTest(TestCase):
    def setUp(self) -> None:
        super().setUp()
        os.environ.pop("SETTINGS_MODULE", None)
        settings.reset()

    def tearDown(self) -> None:
        super().tearDown()
        settings.reset()
        os.environ.pop("SETTINGS_MODULE", None)

    def test_default_config(self):
        self.assertEqual("test", settings.NAME)

    def test_configured(self):
        settings.configure(test_config)
        self.assertEqual("THIS_IS_MY_NAME", settings.NAME)

    def test_priority(self):
        settings.configure(test_config)
        os.environ["SETTINGS_MODULE"] = "../tests/resources/test_config_1.py"
        self.assertEqual("THIS_IS_MY_NAME", settings.NAME)

    def test_double_configure(self):
        settings.configure(test_config)
        with self.assertRaises(AssertionError) as e:
            settings.configure(test_config)
        self.assertIn("Settings are already configured", str(e.exception))

        second_setting = LazySettings()
        with self.assertRaises(AssertionError) as e:
            second_setting.configure(test_config)
        self.assertIn("Do not use multiple settings", str(e.exception))

    def test_environment(self):
        os.environ["SETTINGS_MODULE"] = "tests.resources.configs.test_config_1"
        self.assertEqual("THIS_IS_MY_NAME_TOO", settings.NAME)
        self.assertEqual("bsdf", settings.TEST)
        del os.environ["SETTINGS_MODULE"]

    @override_context(NAME="ABC")
    def test_overwrite(self):
        @override_context(NAME="ASDF")
        def asdf():
            self.assertEqual("ASDF", settings.NAME)
            self.assertTrue(settings._wrapped.is_overridden("NAME"))

        settings.configure(test_config)
        asdf()

        self.assertEqual("ABC", settings.NAME)
        self.assertTrue(settings._wrapped.is_overridden("NAME"))

    def test_context(self):
        settings.configure(test_config)
        self.assertEqual("THIS_IS_MY_NAME", settings.NAME)

        with override_context(NAME="abc"):
            self.assertEqual("abc", settings.NAME)
        self.assertEqual("THIS_IS_MY_NAME", settings.NAME)


@override_settings(BROKER_SUB=f"inproc://test-sub-aaaaaaaaaa")
class ZmqConfigTest(ZMQTestMixin):
    """
    the zmq test mixin dictates the broker pub and sub addresses,
    this test ensures that, in order to avoid accidentially testing on a different protocol,
    because the default settings of a service say so. 
    
    This together with test_test_zeromq.py:ZMQTestMixinInprocTest tests for a bug introduced in
    11a75bf0ae5556d058ae4f1bee4fc1eae5f655e2
    """

    def test_sub(self):
        self.assertNotIn("aaaaaaaaaa", self.zmq_config().BROKER_SUB)



"""datetime models

Revision ID: ff26e54332f9
Revises: 
Create Date: 2021-03-25 11:17:19.954291

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = 'ff26e54332f9'
down_revision = None
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.create_table('test_timezone_invalid',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('datetime', sa.DateTime(), nullable=True),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('test_timezone_invalid')
    # ### end Alembic commands ###



from logging.config import fileConfig

from alembic import context
from sqlalchemy import engine_from_config
from sqlalchemy import pool

from openmodule.checks import check_invalid_database_column_type
from openmodule.database.custom_types import CustomType

# this is the Alembic Config object, which provides
# access to the values within the .ini file in use.

config = context.config

# Interpret the config file for Python logging.
# This line sets up loggers basically.
if config.attributes.get('configure_logging', True):
    fileConfig(config.config_file_name)

target_metadata = config.attributes["target_metadata"]


def render_item(type_, obj, autogen_context):
    """Apply custom rendering for selected items."""

    check_invalid_database_column_type(obj)
    custom_import, name = CustomType.custom_import(obj)
    if type_ == 'type' and custom_import:
        autogen_context.imports.add(f"from {custom_import} import {name}")
        return "%r" % obj
    return False


def run_migrations_online():
    def process_revision_directives(context, revision, directives):
        if config.cmd_opts.autogenerate:
            script = directives[0]
            if script.upgrade_ops.is_empty():
                directives[:] = []

    connectable = config.attributes.get('connection', None)
    if connectable is None:
        connectable = engine_from_config(
            config.get_section(config.config_ini_section),
            prefix="sqlalchemy.",
            poolclass=pool.NullPool,
        )

    with connectable.connect() as connection:
        context.configure(
            connection=connection, target_metadata=target_metadata,
            process_revision_directives=process_revision_directives,
            render_item=render_item,
            render_as_batch=True
        )

        with context.begin_transaction():
            context.run_migrations()


if context.is_offline_mode():
    assert False, "Not in use"
else:
    run_migrations_online()



import logging
import os
import time
from multiprocessing import Process, Event
from signal import SIGINT, SIGKILL
from typing import Callable, Union, Type
from unittest import TestCase

import multiprocessing_logging

from openmodule.core import shutdown_openmodule
from openmodule.utils.schema import Schema
from openmodule_test.health import HealthTestMixin


class ExceptionProcess(Process):
    def __init__(self, *args, **kwargs):
        is_finished = Event()
        super().__init__(*args, **kwargs)
        self.is_finished = is_finished

    def run(self):
        try:
            super().run()
            self.is_finished.set()
        except Exception as e:
            logging.exception(e)


class InterruptTestMixin(TestCase):
    """
    Helper class for testing interrupts and exceptions in code
    for usage, look at file tests/test_interrupt
    """

    def tearDown(self):
        super().tearDown()
        Schema.to_file()

    def wait_for_setup(self):
        pass

    def _wait_for_and_uninstall_mp_handler(self, logger=None):
        """
        waits until the multiprocessing logging handler has finished
        """
        if logger is None:
            logger = logging.getLogger()

        for handler in logger.handlers:
            if isinstance(handler, multiprocessing_logging.MultiProcessingHandler):
                handler.close()

        multiprocessing_logging.uninstall_mp_handler(logger)

    def signal_in_function(self, f: Callable, signal: Union[Type[KeyboardInterrupt], int], *,
                           raise_exception_after: float = 3.0, shutdown_timeout: float = 3.0):
        if signal == KeyboardInterrupt:
            signal = SIGINT
        multiprocessing_logging.install_mp_handler()
        process = ExceptionProcess(target=f)
        process.start()
        self.wait_for_setup()
        time.sleep(raise_exception_after)
        self.assertFalse(process.is_finished.is_set())
        os.kill(process.pid, signal)
        if process.is_finished.wait(timeout=shutdown_timeout):
            self._wait_for_and_uninstall_mp_handler()
            return
        else:
            if process.is_alive():
                os.kill(process.pid, SIGKILL)
                raise TimeoutError("Process took to long for shutdown")
            else:
                raise AssertionError("Process did not finish gracefully")


class MainTestMixin(HealthTestMixin, InterruptTestMixin):
    topics = ["healthz"]
    protocol = "tcp://"

    def wait_for_setup(self):
        self.wait_for_health()

    def tearDown(self):
        try:
            shutdown_openmodule()
        except AssertionError:
            pass
        super().tearDown()



from functools import partial
from typing import Any
from unittest import TestCase

from openmodule.config import settings
from openmodule.dispatcher import MessageDispatcher, SubscribingMessageDispatcher
from openmodule.models.base import ZMQMessage, OpenModuleModel


class MessageDispatcherBaseTest(TestCase):
    dispatcher: MessageDispatcher
    message: Any

    def dummy_message(self, type="test", **kwargs):
        return {"type": type, "name": "testclient", **kwargs}

    def _set_true_handler(self, message, var="message"):
        """test handler"""
        setattr(self, var, message)

    def setUp(self):
        super().setUp()
        self.dispatcher = MessageDispatcher()
        self.message = None

    def tearDown(self) -> None:
        settings.reset()
        super().tearDown()


class MessageDispatcherBasicsTestCase(MessageDispatcherBaseTest):
    def test_topic_is_str(self):
        self.dispatcher.register_handler("test", ZMQMessage, self._set_true_handler)
        self.assertIsNone(self.message)

        self.dispatcher.dispatch(b"test", self.dummy_message())
        self.assertIsNotNone(self.message)

    def test_no_filter(self):
        self.dispatcher.register_handler(b"test", ZMQMessage, self._set_true_handler)
        self.assertIsNone(self.message)

        self.dispatcher.dispatch(b"test", self.dummy_message())
        self.assertIsNotNone(self.message)

    def test_filter(self):
        self.dispatcher.register_handler(b"test", ZMQMessage, self._set_true_handler, filter={"type": "some-type"})

        # filter does not match
        self.dispatcher.dispatch(b"test", self.dummy_message(type="incorrect"))
        self.assertIsNone(self.message)

        # filter matches
        self.dispatcher.dispatch(b"test", self.dummy_message("some-type"))
        self.assertIsNotNone(self.message)

    def test_multiple_filter(self):
        # multiple handlers with identical filters and topics are allowed
        self.message1 = None
        self.message2 = None

        self.dispatcher.register_handler(b"test", ZMQMessage, partial(self._set_true_handler, var="message1"),
                                         filter={"type": "some-type"})
        self.dispatcher.register_handler(b"test", ZMQMessage, partial(self._set_true_handler, var="message2"),
                                         filter={"type": "some-type"})

        # filter does not match
        self.dispatcher.dispatch(b"test", self.dummy_message(type="incorrect"))
        self.assertIsNone(self.message1)
        self.assertIsNone(self.message2)

        # filter matches
        self.dispatcher.dispatch(b"test", self.dummy_message("some-type"))
        self.assertIsNotNone(self.message1)
        self.assertIsNotNone(self.message2)

    def test_prefix_does_not_match(self):
        # since zmq subscribes on all topics as prefixes, the dispatcher
        # has to full-match the topic
        self.dispatcher.register_handler(b"test", ZMQMessage, self._set_true_handler)
        self.dispatcher.dispatch(b"testp", self.dummy_message())
        self.assertIsNone(self.message)

        # exact match works
        self.dispatcher.dispatch(b"test", self.dummy_message())
        self.assertIsNotNone(self.message)

    def test_match_type(self):
        class TypedZMQMessage(ZMQMessage):
            type: str = "test1"

        self.dispatcher.register_handler(b"test", TypedZMQMessage, self._set_true_handler, match_type=True)

        # wrong type does not match
        self.dispatcher.dispatch(b"test", ZMQMessage(type="test2"))
        self.assertIsNone(self.message)

        # correct type, message is dispatched
        self.dispatcher.dispatch(b"test", ZMQMessage(type="test1"))
        self.assertIsNotNone(self.message)

    def test_match_type_without_default_value(self):
        # empty string is also "false" thus not a valid type
        with self.assertRaises(AssertionError) as e:
            self.dispatcher.register_handler(b"test", ZMQMessage, self._set_true_handler, match_type=True)
        self.assertIn("class definition does not set a `type` field", str(e.exception))

        class MessageWithoutType(OpenModuleModel):
            pass

        with self.assertRaises(AssertionError) as e:
            self.dispatcher.register_handler(b"test", MessageWithoutType, self._set_true_handler, match_type=True)
        self.assertIn("class definition does not set a `type` field", str(e.exception))


class MessageDispatcherWithoutExecutorTestCase(MessageDispatcherBaseTest):
    def test_exception_in_handler(self):
        def raises_exception(message):
            raises_exception.register_schema = False
            raise Exception("something broke!")

        self.dispatcher.register_handler(b"test", ZMQMessage, raises_exception, register_schema=False)
        with self.assertLogs() as cm:
            self.dispatcher.dispatch(b"test", self.dummy_message())
        self.assertIn("something broke!", cm.output[0])

    def test_validation_error(self):
        def handler(x):
            pass

        handler.register_schema = False
        self.dispatcher.register_handler("test", ZMQMessage, handler, register_schema=False)
        with self.assertLogs() as cm:
            self.dispatcher.dispatch(b"test", {})
        self.assertIn("validation error", cm.output[0])


class SubscribingMessageDispatcherTestCase(TestCase):
    subscriptions = set()

    def subscribe(self, topic):
        self.subscriptions.add(topic)

    def unsubscribe(self, topic):
        self.subscriptions.remove(topic)

    def setUp(self) -> None:
        self.subscriptions.clear()
        self.dispatcher = SubscribingMessageDispatcher(
            subscribe=self.subscribe,
            unsubscribe=self.unsubscribe
        )

    def test_subscribe(self):
        self.dispatcher.register_handler(b"topic1", ZMQMessage, lambda *x: x, register_schema=False)
        self.assertEqual({b"topic1"}, self.subscriptions)

    def test_unsubscribe(self):
        listener1 = self.dispatcher.register_handler(b"topic1", ZMQMessage, lambda *x: x, register_schema=False)
        listener2 = self.dispatcher.register_handler(b"topic1", ZMQMessage, lambda *x: x, register_schema=False)

        self.assertEqual({b"topic1"}, self.subscriptions)

        # listener 2 is still subscribed, topic should still be listed in subscriptions
        self.dispatcher.unregister_handler(listener1)
        self.assertEqual({b"topic1"}, self.subscriptions)

        # now no listener is subscribed anymore
        self.dispatcher.unregister_handler(listener2)
        self.assertEqual(set(), self.subscriptions)



import logging
import time
from signal import SIGINT, SIGTERM, SIGABRT
from unittest import TestCase

from openmodule.config import settings
from openmodule.core import init_openmodule, shutdown_openmodule
from openmodule_test.interrupt import InterruptTestMixin, MainTestMixin


def test():
    try:
        test.run = True
        t0 = time.time()
        while test.run and time.time() - t0 < 10:
            time.sleep(1)
    except KeyboardInterrupt:
        logging.error("Keyboard")
    except Exception as e:
        logging.info(f"exception: {e}")
    else:
        logging.info("not running")
    finally:
        logging.info("killed")


class InterruptTest(InterruptTestMixin, TestCase):
    def test_keyboard(self):
        with self.assertLogs() as cm:
            self.signal_in_function(test, KeyboardInterrupt)
        self.assertIn("Keyboard", str(cm.output))

    def test_sigterm(self):
        with self.assertRaises(AssertionError) as e:
            self.signal_in_function(test, SIGTERM)
        self.assertIn("Process did not finish gracefully", str(e.exception))


def sleepy():
    try:
        t0 = time.time()
        while time.time() - t0 < 10:
            time.sleep(1)
    except:
        try:
            sleepy()
        except SystemExit:
            pass


class InterruptTimeoutTest(InterruptTestMixin, TestCase):
    function = staticmethod(sleepy)

    def test_exception(self):
        with self.assertRaises(TimeoutError) as e:
            self.signal_in_function(sleepy, KeyboardInterrupt)
        self.assertIn("Process took to long for shutdown", str(e.exception))


def main():
    core = init_openmodule(settings)
    main.run = True
    try:
        while main.run:
            time.sleep(1)
        logging.info("stopped")
    except KeyboardInterrupt:
        logging.error("keyboard")
    except Exception as e:
        print(e)
    finally:
        shutdown_openmodule()


def main_without_sighandler():
    core = init_openmodule(settings, catch_sigterm=False)
    main.run = True
    try:
        while main.run:
            time.sleep(1)
        logging.info("stopped")
    except KeyboardInterrupt:
        logging.error("keyboard")
    except Exception as e:
        print(e)
    finally:
        shutdown_openmodule()


class MainTest(MainTestMixin):
    def test_keyboard(self):
        with self.assertLogs() as cm:
            self.signal_in_function(main, SIGINT)
        self.assertIn("keyboard", str(cm.output))

    def test_docker_shutdown(self):
        with self.assertLogs() as cm:
            self.signal_in_function(main, SIGTERM)
        self.assertIn("keyboard", str(cm.output))

    def test_docker_uncaught_shutdown(self):
        with self.assertRaises(AssertionError) as e:
            self.signal_in_function(main_without_sighandler, SIGTERM)
        self.assertIn("Process did not finish gracefully", str(e.exception))


def exception_raiser():
    core = init_openmodule(settings, catch_sigterm=False)
    main.run = True
    try:
        while main.run:
            time.sleep(1)
            break_dict = dict()
            not_working = break_dict["breaking"]
        logging.info("stopped")
    except KeyboardInterrupt:
        logging.error("keyboard")
    finally:
        shutdown_openmodule()


class ExceptionInFunctionTest(MainTestMixin):
    def test_uncaught_exception_in_function(self):
        with self.assertRaises(AssertionError) as e:
            self.signal_in_function(exception_raiser, KeyboardInterrupt)
        self.assertIn("Process did not finish gracefully", str(e.exception))




import os
import time
from threading import Thread
from unittest import TestCase

from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm.exc import NoResultFound, MultipleResultsFound, DetachedInstanceError

from openmodule.config import settings
from openmodule.database.database import Database, active_databases, database_path
from openmodule.utils.db_helper import update_query, delete_query
from openmodule_test.database import SQLiteTestMixin
from openmodule_test.database_models import DatabaseTestModel


class DatabaseTest(SQLiteTestMixin, TestCase):
    alembic_path = "../tests/test_database_data"

    def test_filter(self):
        data = [DatabaseTestModel(value1=x, value2=x % 3) for x in range(5)]
        with self.database as db:
            db.add_all(data)
            self.assertEqual(1, db.query(DatabaseTestModel).filter_by(value1=0).count())
            self.assertEqual(2, db.query(DatabaseTestModel).filter_by(value2=0).count())

        with self.database as db:
            query = db.query(DatabaseTestModel).filter_by(value2=0)
            res = query.filter_by(value1=0)
            self.assertEqual(1, len(list(res)))
            self.assertEqual(0, res[0].value1)
            self.assertEqual(0, res[0].value2)

    def test_update_filter(self):
        data = [DatabaseTestModel(value1=x, value2=x % 3) for x in range(5)]
        with self.database as db:
            db.add_all(data)

            rows_changed = update_query(db, db.query(DatabaseTestModel).filter_by(value2=2), dict(value1=7))
            self.assertEqual(rows_changed, 1)
            results = db.query(DatabaseTestModel).filter_by(value2=2)
            self.assertEqual(2, results[0].value2)

            rows_changed = update_query(db, db.query(DatabaseTestModel).filter_by(value2=1), dict(value1=7))
            self.assertEqual(rows_changed, 2)
            results = db.query(DatabaseTestModel).filter_by(value2=1)
            self.assertEqual(2, results.count())
            self.assertEqual(3, db.query(DatabaseTestModel).filter_by(value1=7).count())

    def test_expire_on_commit(self):
        with self.database as db:
            model = DatabaseTestModel(value1=1)
            db.add(model)
            db.flush()
            pk = model.id

        # default expire is True
        with self.database as db:
            model = db.query(DatabaseTestModel).get(pk)
        with self.assertRaises(DetachedInstanceError):
            self.assertEqual(model.value1, 1)

        # default expire also is True when calling database
        with self.database() as db:
            model = db.query(DatabaseTestModel).get(pk)
        with self.assertRaises(DetachedInstanceError):
            self.assertEqual(model.value1, 1)

        # passing expire_on_commit disables the expire
        with self.database(expire_on_commit=False) as db:
            model = db.query(DatabaseTestModel).get(pk)
        self.assertEqual(model.value1, 1)

    def test_delete(self):
        data = [DatabaseTestModel(value1=x % 5, value2=x % 3) for x in range(10)]
        with self.database as db:
            db.add_all(data)

            delete_query(db, db.query(DatabaseTestModel).filter_by(value2=0, value1=17))
            self.assertEqual(10, len(db.query(DatabaseTestModel).all()))

            nr_deleted = delete_query(db, db.query(DatabaseTestModel).filter_by(value2=0))
            self.assertEqual(6, len(db.query(DatabaseTestModel).all()))
            self.assertEqual(4, nr_deleted)

            nr_deleted = delete_query(db, db.query(DatabaseTestModel).filter_by(value2=1, value1=1))
            self.assertEqual(5, len(db.query(DatabaseTestModel).all()))
            self.assertEqual(1, nr_deleted)

            tmp = db.query(DatabaseTestModel).filter_by(value1=4)

            self.assertTrue(1, tmp.count())
            db.delete(tmp[0])
            self.assertFalse(any(x.value1 == 4 for x in db.query(DatabaseTestModel).all()))

    def test_rollback(self):
        a = DatabaseTestModel(value1=5, value2=0)
        with self.database as db:
            db.add(a)
            db.flush()
            tmp = a.id
        with self.assertRaises(Exception):
            with self.database as db:
                a.value1 = 6
                db.add(a)
                raise Exception("asdf")

        with self.database as db:
            current = db.query(DatabaseTestModel).get(tmp)
            self.assertEqual(5, current.value1)
            self.assertEqual(1, len(db.query(DatabaseTestModel).all()))

    def test_exception_on_commit(self):
        with self.database as db:
            db.add(DatabaseTestModel(id="id1", value1=1))

        with self.assertRaises(IntegrityError) as e:
            with self.database as db:  # this add will fail, because id1 is already in use
                db.add(DatabaseTestModel(id="id1", value1=1))
        self.assertIn("UNIQUE constraint", str(e.exception))

        # test that we are able to use the database afterwards
        with self.database as db:
            db.add(DatabaseTestModel(id="id2", value1=2))

    def test_database_transaction_is_single_threaded(self):
        """
        this test starts a second thread, which waits for 0.5s, after that it tries to acquire the database
        and read the value of an object. The main thread in the meantime blocks for 2 seconds, and after that
        writes something to the database. Both threads check that the other is at the expected state.
        The test shows that the transaction is single threaded, and all threads are blocked while
        a thread has acquired the db
        """
        main_thread_position = ""
        second_thread_position = ""
        second_thread_value = None

        # initial setup model(1) with value = initial
        with self.database as db:
            model = DatabaseTestModel(id="1", string="initial")
            db.add(model)

        def second_thread():
            nonlocal second_thread_position, second_thread_value
            time.sleep(0.5)
            second_thread_position = "acquiring"
            with self.database as db:
                self.assertEqual(main_thread_position, "end")
                second_thread_model: DatabaseTestModel = db.query(DatabaseTestModel).get("1")
                second_thread_value = second_thread_model.string

        thread = Thread(target=second_thread)
        thread.start()

        with self.database as db:
            time.sleep(2)
            self.assertEqual(second_thread_position, "acquiring")
            model = db.query(DatabaseTestModel).get("1")
            model.string = "changed"
            db.add(model)
            main_thread_position = "end"

        thread.join()
        self.assertEqual(second_thread_value, "changed")

    def test_multiple_databases(self):
        db1 = Database(self.database_folder, name="asdf", alembic_path=self.alembic_path)
        with db1 as db:
            db.add(DatabaseTestModel(value1=0, value2=4))
        with self.assertRaises(Exception) as e:
            db2 = Database(self.database_folder, name="asdf", alembic_path=self.alembic_path)
        self.assertIn("already exists", str(e.exception))
        self.assertTrue(os.path.exists(os.path.join(self.database_folder, "asdf.sqlite3")))

    def test_get(self):
        with self.database as db:
            base = db.query(DatabaseTestModel)
            res = base.get("asdf")
            self.assertIsNone(res)

            with self.assertRaises(Exception) as e:
                base.one()
            self.assertEqual(NoResultFound, type(e.exception))

            res = base.one_or_none()
            self.assertIsNone(res)

            res = base.first()
            self.assertIsNone(res)

            model1 = DatabaseTestModel()
            model2 = DatabaseTestModel()
            db.add(model1)
            db.add(model2)

            res = base.first()
            self.assertIsNotNone(res)

            with self.assertRaises(Exception) as e:
                base.one_or_none()
            self.assertEqual(MultipleResultsFound, type(e.exception))

    def test_closed_session(self):
        model = DatabaseTestModel()

        with self.database as db:
            db.add(model)

        with self.assertRaises(AssertionError) as e:
            db.query(DatabaseTestModel).all()
        self.assertIn("Session is already closed", str(e.exception))

    def test_closed_session_write(self):
        model = DatabaseTestModel()
        model1 = DatabaseTestModel()

        with self.database as db:
            db.add(model)

        with self.assertRaises(AssertionError) as e:
            db.add(model1)
        self.assertIn("Session is already closed", str(e.exception))


class ShutdownTestCase(TestCase):
    alembic_path = "../tests/test_database_data"

    def get_database(self):
        return Database(settings.DATABASE_FOLDER, name="shutdown", alembic_path=self.alembic_path)

    def check_db_present(self, value):
        tmp = database_path(settings.DATABASE_FOLDER, "shutdown")
        if value:
            self.assertIn(tmp, active_databases.keys())
        else:
            self.assertNotIn(tmp, active_databases.keys())

    def test_shutdown(self):
        self.check_db_present(False)
        database = self.get_database()
        self.check_db_present(True)
        database.shutdown()
        self.check_db_present(False)

    def test_shutdown_while_using(self):
        database = self.get_database()

        def hold_db():
            with database as db:
                time.sleep(3)

        thread = Thread(target=hold_db)
        thread.start()
        self.assertEqual(True, thread.is_alive())
        self.check_db_present(True)
        database.shutdown()
        self.assertEqual(False, thread.is_alive())
        self.check_db_present(False)

    def test_double_shutdown(self):
        database = self.get_database()
        self.check_db_present(True)

        database.shutdown()

        with self.assertRaises(AssertionError) as e:
            database.shutdown()
        self.assertIn("already closed", str(e.exception))
        self.check_db_present(False)


class MultipleMigrationTest(SQLiteTestMixin, TestCase):
    """
    there was an issue, which prevented a database from beeing migrated multiple times in a single process
    """
    create_database = False
    alembic_path = "../tests/test_database_data"

    def test_migrate_1(self):
        database = Database(self.database_folder, self.database_name, self.alembic_path)
        with database as db:
            self.assertEqual(0, db.query(DatabaseTestModel).count())
        database.shutdown()
        self.delete_database(database)

    def test_migrate_2(self):
        database = Database(self.database_folder, self.database_name, self.alembic_path)
        with database as db:
            self.assertEqual(0, db.query(DatabaseTestModel).count())
        database.shutdown()
        self.delete_database(database)



from datetime import datetime
from unittest import TestCase

import freezegun
from dateutil.parser import parse

from openmodule.config import settings
from openmodule.models.backend import Access, AccessCategory


class AccessModelTestCase(TestCase):
    def assertTzNaive(self, d):
        self.assertTrue(d.tzinfo is None)

    def tearDown(self) -> None:
        super().tearDown()
        settings.reset()

    def is_valid_at(self, access, dt, timezone="Europe/Vienna"):
        dt = parse(dt)
        return access.is_valid_at(dt, timezone)

    def test_deserialized_datetimes_are_timezone_native(self):
        # unix timestamps
        json_bytes = b'{"start": 1600000, "user": "some-user", "category": "booked-digimon"}'
        access = Access.parse_raw(json_bytes)
        self.assertTzNaive(access.start)

        # tz aware iso strings
        json_bytes = b'{"start": "2017-02-15T20:26:08.937881-06:00", "user": "some-user", "category": "booked-digimon"}'
        access = Access.parse_raw(json_bytes)
        self.assertTzNaive(access.start)

        # naive iso strings
        json_bytes = b'{"start": "2017-02-15T20:26:08.937881", "user": "some-user", "category": "booked-digimon"}'
        access = Access.parse_raw(json_bytes)
        self.assertTzNaive(access.start)

    def test_is_valid_non_recurrent_start_end(self):
        access = Access(start="2000-01-02T00:00", end="2000-01-03T00:00", user="test", category="booked-digimon")

        # naive test
        self.assertFalse(self.is_valid_at(access, "2000-01-01T23:59:59"))
        self.assertTrue(self.is_valid_at(access, "2000-01-02T00:00"))
        self.assertTrue(self.is_valid_at(access, "2000-01-03T00:00"))
        self.assertFalse(self.is_valid_at(access, "2000-01-03T00:00:01"))

        # timezone aware test
        self.assertFalse(self.is_valid_at(access, "2000-01-02T01:59+02:00"))
        self.assertTrue(self.is_valid_at(access, "2000-01-02T02:00+02:00"))
        self.assertTrue(self.is_valid_at(access, "2000-01-03T02:00+02:00"))
        self.assertFalse(self.is_valid_at(access, "2000-01-03T02:01+02:00"))

    def test_is_valid_timezone_issue(self):
        access = Access(user="b1", start="2020-01-01 00:00", end="2020-01-01 04:00",
                        category=AccessCategory.booked_digimon)

        with freezegun.freeze_time("2020-01-01 03:30"):
            access = access.parse_raw(access.json_bytes())
            self.assertTrue(access.is_valid_at(datetime.utcnow(), None))

    def test_is_valid_non_recurrent_start_no_end(self):
        access = Access(start="2000-01-02T00:00", user="test", category="booked-digimon")

        # naive test
        self.assertFalse(self.is_valid_at(access, "2000-01-01T23:59:59"))
        self.assertTrue(self.is_valid_at(access, "2000-01-02T00:00"))
        self.assertTrue(self.is_valid_at(access, "3000-02-02T00:00"))

        # timezone aware test
        self.assertFalse(self.is_valid_at(access, "2000-01-01T01:59:59+02:00"))
        self.assertTrue(self.is_valid_at(access, "2000-01-02T02:00+02:00"))
        self.assertTrue(self.is_valid_at(access, "3000-02-02T02:00+02:00"))

    def test_is_valid_recurrent_start_end_recurrent(self):
        access = Access(start="2000-01-01T00:00", end="2000-01-07T23:59", user="test", category="booked-visitor",
                        recurrence="DTSTART:19990108T110000\nRRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR",
                        duration=3600 * 9)  # from 12:00 to 21:00

        # before the start date
        self.assertFalse(self.is_valid_at(access, "1999-12-30T13:00"))  # SA
        self.assertFalse(self.is_valid_at(access, "1999-12-31T13:00"))  # FR

        # between start and end
        self.assertFalse(self.is_valid_at(access, "2000-01-01T13:00"))  # SA
        self.assertFalse(self.is_valid_at(access, "2000-01-02T13:00"))  # SO
        self.assertTrue(self.is_valid_at(access, "2000-01-03T13:00"))  # MO
        self.assertTrue(self.is_valid_at(access, "2000-01-04T13:00"))  # TU
        self.assertTrue(self.is_valid_at(access, "2000-01-05T13:00"))  # WE
        self.assertTrue(self.is_valid_at(access, "2000-01-06T13:00"))  # TH
        self.assertTrue(self.is_valid_at(access, "2000-01-07T13:00"))  # FR

        # Test the exact limits on friday
        self.assertFalse(self.is_valid_at(access, "2000-01-07T11:59:59+01:00"))  # FR
        self.assertTrue(self.is_valid_at(access, "2000-01-07T12:00+01:00"))  # FR
        self.assertTrue(self.is_valid_at(access, "2000-01-07T20:59:59+01:00"))  # FR
        self.assertFalse(self.is_valid_at(access, "2000-01-07T21:00:00+01:00"))  # FR

        # after the end date
        self.assertFalse(self.is_valid_at(access, "2000-01-08T13:00"))  # SA
        self.assertFalse(self.is_valid_at(access, "2000-01-09T13:00"))  # FR

    def test_is_valid_recurrent_start_no_end_recurrent(self):
        access = Access(start="2000-01-01T00:00", end=None, user="test", category="booked-visitor",
                        recurrence="DTSTART:19990108T110000\nRRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR",
                        duration=3600 * 9)  # from 12:00 to 21:00

        # before the start date
        self.assertFalse(self.is_valid_at(access, "1999-12-30T13:00"))  # SA
        self.assertFalse(self.is_valid_at(access, "1999-12-31T13:00"))  # FR

        # between start and end
        self.assertFalse(self.is_valid_at(access, "2000-01-01T13:00"))  # SA
        self.assertFalse(self.is_valid_at(access, "2000-01-02T13:00"))  # SO
        self.assertTrue(self.is_valid_at(access, "2000-01-03T13:00"))  # MO
        self.assertTrue(self.is_valid_at(access, "2000-01-04T13:00"))  # TU
        self.assertTrue(self.is_valid_at(access, "2000-01-05T13:00"))  # WE
        self.assertTrue(self.is_valid_at(access, "2000-01-06T13:00"))  # TH
        self.assertTrue(self.is_valid_at(access, "2000-01-07T13:00"))  # FR

        # years later
        self.assertFalse(self.is_valid_at(access, "2022-01-01T13:00"))  # SA
        self.assertFalse(self.is_valid_at(access, "2022-01-02T13:00"))  # SO
        self.assertTrue(self.is_valid_at(access, "2022-01-03T13:00"))  # MO
        self.assertTrue(self.is_valid_at(access, "2022-01-04T13:00"))  # TU
        self.assertTrue(self.is_valid_at(access, "2022-01-05T13:00"))  # WE
        self.assertTrue(self.is_valid_at(access, "2022-01-06T13:00"))  # TH
        self.assertTrue(self.is_valid_at(access, "2022-01-07T13:00"))  # FR

    def test_recurrence_during_dst_change(self):
        access = Access(start="2000-01-01T00:00", end=None, user="test", category="booked-visitor",
                        recurrence="DTSTART:19990108T000000\nRRULE:FREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR,SA,SU",
                        duration=3600 * 3 + 1)  # from 01:00 to 04:00

        # normal day
        self.assertFalse(self.is_valid_at(access, "2021-03-27T00:00+01:00"))
        self.assertTrue(self.is_valid_at(access, "2021-03-27T01:00+01:00"))
        self.assertTrue(self.is_valid_at(access, "2021-03-27T01:59+01:00"))
        self.assertTrue(self.is_valid_at(access, "2021-03-27T03:00+01:00"))
        self.assertTrue(self.is_valid_at(access, "2021-03-27T04:00+01:00"))
        self.assertFalse(self.is_valid_at(access, "2021-03-27T05:00+01:00"))  # this is the normal case
        self.assertFalse(self.is_valid_at(access, "2021-03-27T06:00+01:00"))

        # from +01:00 to +02:00
        self.assertFalse(self.is_valid_at(access, "2021-03-28T00:00+01:00"))
        self.assertTrue(self.is_valid_at(access, "2021-03-28T01:00+01:00"))
        self.assertTrue(self.is_valid_at(access, "2021-03-28T01:59+01:00"))
        self.assertTrue(self.is_valid_at(access, "2021-03-28T03:00+02:00"))
        self.assertTrue(self.is_valid_at(access, "2021-03-28T04:00+02:00"))
        # usually it goes only until 4 o'clock, but during dst change we add the hour to avoid conflicts
        self.assertTrue(self.is_valid_at(access, "2021-03-28T05:00+02:00"))  # special case
        self.assertFalse(self.is_valid_at(access, "2021-03-28T06:00+02:00"))

        # from +02:00 to +01:00
        self.assertFalse(self.is_valid_at(access, "2021-10-31T00:00+02:00"))
        self.assertTrue(self.is_valid_at(access, "2021-10-31T01:00+02:00"))
        self.assertTrue(self.is_valid_at(access, "2021-10-31T01:59+02:00"))
        self.assertTrue(self.is_valid_at(access, "2021-10-31T03:00+01:00"))
        self.assertTrue(self.is_valid_at(access, "2021-10-31T04:00+01:00"))
        # usually it goes only until 4 o'clock, but during dst change we add the hour to avoid conflicts
        self.assertTrue(self.is_valid_at(access, "2021-10-31T05:00+01:00"))  # special case
        self.assertFalse(self.is_valid_at(access, "2021-10-31T06:00+01:00"))



import requests
import requests_mock

from openmodule_test.utils import DeveloperError


class ApiMocker:
    host = None

    def __init__(self, mocker):
        self.mocker = mocker
        if not self.host:
            raise DeveloperError("you need to set a host")

    def server_url(self, url):
        return f"{self.host.rstrip('/')}/{url}"

    def ok(self):
        self.mocker.get(requests_mock.ANY, status_code=200, json={})
        self.mocker.post(requests_mock.ANY, status_code=200, json={})
        self.mocker.put(requests_mock.ANY, status_code=200, json={})
        self.mocker.delete(requests_mock.ANY, status_code=200, json={})

    def timeout(self):
        self.mocker.get(requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)
        self.mocker.post(requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)
        self.mocker.put(requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)
        self.mocker.delete(requests_mock.ANY, exc=requests.exceptions.ConnectTimeout)

    def unavailable(self):
        self.mocker.get(requests_mock.ANY, status_code=503, json={})
        self.mocker.post(requests_mock.ANY, status_code=503, json={})
        self.mocker.put(requests_mock.ANY, status_code=503, json={})
        self.mocker.delete(requests_mock.ANY, status_code=503, json={})

    def server_error(self):
        self.mocker.get(requests_mock.ANY, status_code=500, json={})
        self.mocker.post(requests_mock.ANY, status_code=500, json={})
        self.mocker.put(requests_mock.ANY, status_code=500, json={})
        self.mocker.delete(requests_mock.ANY, status_code=500, json={})

    def invalid_auth(self):
        self.mocker.get(requests_mock.ANY, status_code=403, json={})
        self.mocker.post(requests_mock.ANY, status_code=403, json={})
        self.mocker.get(requests_mock.ANY, status_code=403, json={})
        self.mocker.post(requests_mock.ANY, status_code=403, json={})

    def bad_request(self):
        self.mocker.get(requests_mock.ANY, status_code=400, json={})
        self.mocker.post(requests_mock.ANY, status_code=400, json={})
        self.mocker.get(requests_mock.ANY, status_code=400, json={})
        self.mocker.post(requests_mock.ANY, status_code=400, json={})

    def not_found(self):
        self.mocker.get(requests_mock.ANY, status_code=404, json={})
        self.mocker.post(requests_mock.ANY, status_code=404, json={})
        self.mocker.get(requests_mock.ANY, status_code=404, json={})
        self.mocker.post(requests_mock.ANY, status_code=404, json={})



class DeveloperError(Exception):
    """
    An exception which indicates a developer error. This is used instead of assertions in testcases because if
    one were to check for AssertionErrors and an assertion is raised not because of a tested condition, but
    because the test-utils were used incorrectly, then there is no way to distinguish between those.

    Use only in testcases, use normal assertions for service code. you can still use assertions in
    setUp(), tearDown(), setUpClass(), tearDownClass() code, because nobody would catch assertions there
    """



from sqlalchemy import DateTime
from sqlalchemy.sql.visitors import VisitableType
from sqlalchemy.types import TypeDecorator


class MetaOptions(VisitableType):
    """
    this behaves similarly to django's meta classes, but without any edge cases covered
    it is used for simple single inheritance without anything special in the alert type
    and package info classes
    """

    def __new__(cls, name, bases, dct):
        x = super().__new__(cls, name, bases, dct)

        # EventModel
        if len(bases) == 1:
            if not getattr(bases[0], "_registry", None):
                setattr(bases[0], "_registry", dict())

            bases[0]._registry[x] = x.__module__
        return x


class CustomType(TypeDecorator, metaclass=MetaOptions):
    @classmethod
    def custom_import(cls, obj):
        for custom, mod in cls._registry.items():
            if type(obj) == custom:
                return mod, custom.__name__
        return None, None


class TZDateTime(CustomType):
    impl = DateTime

    def process_bind_param(self, value, dialect):
        if value is not None:
            assert value.tzinfo is None, "You need to convert a datetime to a naive time, because sqlite loses tz infos"
        return value



from datetime import datetime
from unittest import TestCase

from dateutil.tz import gettz, tzutc
from sqlalchemy.exc import StatementError

from openmodule.config import database_folder
from openmodule.database.database import Database
from openmodule.models.base import OpenModuleModel, timezone_validator
from openmodule_test.database import SQLiteTestMixin
from openmodule_test.database_models import DatabaseTimezoneTestModel


class TimezoneCheckTest(TestCase):
    def test_validator_is_enforced(self):
        with self.assertRaises(AssertionError) as e:
            class MyModel(OpenModuleModel):
                some_field: datetime

        self.assertIn("timezone_validator", str(e.exception))

    def test_datetime_fields_are_converted_to_utc(self):
        class MyModel(OpenModuleModel):
            field: datetime
            _tz_field = timezone_validator("field")

        self.assertEqual("2021-03-25T08:19:42", MyModel(field=1616660382).field.isoformat())
        self.assertEqual("2021-03-25T08:19:42", MyModel(field="2021-03-25 08:19:42").field.isoformat())
        self.assertEqual("2021-03-25T08:19:42", MyModel(field="2021-03-25T08:19:42").field.isoformat())
        self.assertEqual("2021-03-25T08:19:42", MyModel(field="2021-03-25T08:19:42+00:00").field.isoformat())
        self.assertEqual("2021-03-25T08:19:42", MyModel(field="2021-03-25T09:19:42+01:00").field.isoformat())

        self.assertIsNone(MyModel(field="2021-03-25T08:19:42+00:00").field.tzinfo)
        self.assertIsNone(MyModel(field="2021-03-25T09:19:42+01:00").field.tzinfo)

    def test_no_datetime_fields(self):
        with self.assertRaises(AssertionError) as e:
            Database(database_folder(), "test_database", "../tests/invalid_database")
        self.assertIn("Do NOT use DateTime fields, use TZDateTime fields instead", str(e.exception))


class DatabaseDatetimeTest(SQLiteTestMixin):
    alembic_path = "../tests/test_database_data"

    def test_sqlite_timezone(self):
        with self.assertRaises(StatementError) as e:
            with self.database as db:
                now = datetime.now(tzutc()).astimezone(gettz("Europe/Vienna"))
                model = DatabaseTimezoneTestModel(tz_datetime=now)
                db.add(model)
        self.assertIn("You need to convert a datetime to a naive time", str(e.exception))

        with self.database as db:
            now = datetime.utcnow()
            model = DatabaseTimezoneTestModel(tz_datetime=now)
            db.add(model)



import time
from threading import Thread
from unittest import TestCase

from openmodule.models.base import ZMQMessage, Gateway, Direction
from openmodule_test.gate import TestGate
from openmodule_test.utils import DeveloperError
from openmodule_test.zeromq import ZMQTestMixin


class DummyAccessMessage(ZMQMessage):
    name = "test"
    type = "access"
    gateway: Gateway


class TestGateTestCase(ZMQTestMixin, TestCase):
    topics = [b"access_accept", b"access_reject"]

    def setUp(self):
        super().setUp()

        self.gate1 = TestGate("gate1", Direction.IN, self.zmq_client.zmq_dispatcher)
        self.gate2 = TestGate("gate2", Direction.OUT, self.zmq_client.zmq_dispatcher)

    def send_access_accept(self, gate: str, direction: str):
        self.zmq_client.send("access_accept", DummyAccessMessage(gateway=Gateway(gate=gate, direction=direction)))

    def send_access_reject(self, gate: str, direction: str):
        self.zmq_client.send("access_reject", DummyAccessMessage(gateway=Gateway(gate=gate, direction=direction)))

    def test_wait_for_open(self):
        self.send_access_accept("gate1", "in")
        with self.assertRaises(TimeoutError):
            self.gate2.wait_for_open()
        self.gate1.wait_for_open()

    def test_wait_for_reject(self):
        self.send_access_reject("gate1", "in")
        with self.assertRaises(TimeoutError):
            self.gate2.wait_for_reject()
        self.gate1.wait_for_reject()

    def test_first_reject_then_open(self):
        self.send_access_reject("gate1", "in")

        def _send_accept_in_1s():
            time.sleep(1)
            self.send_access_accept("gate1", "in")

        thread = Thread(target=_send_accept_in_1s())
        thread.start()

        with self.assertRaises(AssertionError):
            # in this case an assertion instead of a timeout is raised, because the gate way opened, while
            # we expected a reject
            self.gate1.wait_for_reject(timeout=5)
        with self.assertRaises(TimeoutError):
            self.gate2.wait_for_reject()

        # gate 1 is rejected, but it opened, so assert opened is still true
        with self.assertRaises(AssertionError):
            self.gate1.assert_rejected()
        self.gate1.assert_opened()

        # gate 2 received no messages, so everything asserts
        with self.assertRaises(AssertionError):
            self.gate2.assert_rejected()
        with self.assertRaises(AssertionError):
            self.gate2.assert_opened()

        thread.join()

    def test_reset_counts_required(self):
        self.send_access_accept("gate1", "in")
        self.gate1.wait_for_open()

        with self.assertRaises(DeveloperError) as e:
            self.gate1.wait_for_open()

        self.assertIn("reset_counts()", str(e.exception))



import json

import os
import time
from datetime import datetime, timedelta
from typing import Optional
from unittest import TestCase

import freezegun
import orjson
from pydantic import ValidationError

from openmodule.models.backend import Access
from openmodule.models.base import OpenModuleModel, ZMQMessage, Gateway


class TestModel(OpenModuleModel):
    __test__ = False
    value: str


class ModelTestCase(TestCase):
    def test_json_is_forbidden(self):
        instance = TestModel(value="Hello World")
        with self.assertRaises(AssertionError):
            instance.json()

    def test_json_dumps_and_load(self):
        instance = TestModel(value="Hello World")
        instance2 = TestModel.parse_raw(instance.json_bytes())
        self.assertEqual(instance.value, instance2.value)

    def test_unset_default_are_serialized(self):
        class SomeZMQMessage(ZMQMessage):
            type: str = "my-type"

        msg = SomeZMQMessage(name="myname").json_bytes()
        data = orjson.loads(msg)

        self.assertIn("type", data)
        self.assertEqual("my-type", data.get("type"))

    def test_none_values_are_not_exported(self):
        class SomeZMQMessage(ZMQMessage):
            type: str = "test"
            value: Optional[float]

        data = orjson.loads(SomeZMQMessage(name="myname").json_bytes())
        self.assertNotIn("value", data)

        data = orjson.loads(SomeZMQMessage(name="myname", value=1).json_bytes())
        self.assertIn("value", data)
        self.assertEqual(1, data["value"])

        # to my understanding pydantic, openapi and jsonschema all do not agree how nullable but required
        # fields should correctly be handled / annotated. Currently our strategy is to use exclude_none=True
        # so null values are not exported, in an ideal world i would like to have value=None exported, but
        # it is currently not possible
        data = orjson.loads(SomeZMQMessage(name="myname", value=None).json_bytes())
        # self.assertIn("value", data)
        # self.assertEqual(None, data["value"])
        self.assertNotIn("value", data)


class ZMQMessageTestCase(TestCase):
    def test_timestamp_to_datetime(self):
        instance = ZMQMessage(timestamp=1607586354631, name="test", type="test")
        self.assertEqual(instance.timestamp.isoformat(), "2020-12-10T07:45:54.631000")

    def test_timestamp_is_now_if_unset(self):
        instance = ZMQMessage(name="test", type="test")
        self.assertAlmostEqual(instance.timestamp, datetime.utcnow(), delta=timedelta(seconds=1))

    def test_datetime_is_serialized_to_unix_timestamp(self):
        with freezegun.freeze_time("2020-01-01 00:00"):
            instance = ZMQMessage(name="test", type="test")
            json = instance.json_bytes()
            loaded_json = orjson.loads(json)
            self.assertIsInstance(loaded_json["timestamp"], float)
            self.assertEqual(loaded_json["timestamp"], 1577836800)

        instance = ZMQMessage(name="test", type="test")
        instance.timestamp = datetime(2020, 1, 1, 0, 0, 0, 0)
        self.assertEqual(instance.dict()["timestamp"], 1577836800)

    def test_receive_optional_gateway(self):
        class OptionalGateway(ZMQMessage):
            type = "something"
            gateway: Gateway = Gateway()

        message = OptionalGateway(name="test")
        self.assertEqual(message.gateway.direction, "")
        self.assertEqual(message.gateway.gate, "")

        message = OptionalGateway(name="test", gateway={"direction": "in"})
        self.assertEqual(message.gateway.direction, "in")
        self.assertEqual(message.gateway.gate, "")

        with self.assertRaises(ValidationError) as e:
            OptionalGateway(name="test", gateway={"direction": "incorrect"})

        errors = e.exception.errors()
        self.assertEqual(errors[0]["loc"], ("gateway", "direction"))

    def test_unicode_decode_with_old_omutils(self):
        msg = ZMQMessage(name="tä\nst", type="tö\tst")

        data = msg.json_bytes()
        self.assertTrue(isinstance(data, bytes))

        # ensure no unicode symbols are present
        as_unicode = data.decode()
        as_unicode.encode("ascii")  # check its ascii only
        self.assertNotIn("ä", as_unicode)
        self.assertNotIn("ö", as_unicode)
        self.assertIn("\\t", as_unicode)

        # ensure orjson and legacy json return the same result
        res1 = json.loads(as_unicode)
        res2 = json.loads(data)
        res3 = orjson.loads(as_unicode)
        res4 = orjson.loads(data)

        self.assertEqual(res1, res2)
        self.assertEqual(res2, res3)
        self.assertEqual(res3, res4)

        res1.pop("timestamp")
        self.assertDictEqual(res1, {"name": "tä\nst", "type": "tö\tst"})





from unittest import TestCase

from openmodule.alert import AlertHandler, AlertHandleType
from openmodule.core import init_openmodule, shutdown_openmodule
from openmodule_test.alert import AlertTestMixin


class AlertTestCase(AlertTestMixin, TestCase):
    topics = ["alert"]

    def setUp(self):
        super().setUp()
        self.core = init_openmodule(context=self.zmq_context(), config=self.zmq_config())
        self.addCleanup(shutdown_openmodule)

    def tearDown(self):
        super().tearDown()

    def test_value_is_required_for_state(self):
        alert_handler = AlertHandler(self.core)
        alert_handler.send("some_alert", AlertHandleType.state, value=10, source="pin1")
        self.assertAlert()

        with self.assertRaises(ValueError) as e:
            alert_handler.send("some_alert", AlertHandleType.state, source="pin1")
        self.assertIn("value must not be none", str(e.exception).lower())
        self.assertNoAlert()

    def test_value_none_is_not_sent(self):
        alert_handler = AlertHandler(self.core)
        alert_handler.send("some_alert", AlertHandleType.state_change, source="pin1")
        alert = self.assertAlert(alert_type="some_alert")
        self.assertNotIn("value", alert)



import logging
import threading
import time
from typing import Dict, Optional
from uuid import uuid4

from openmodule.models.rpc import RPCResponse, RPCRequest, RPCResult

from openmodule.core import OpenModuleCore
from openmodule.rpc.common import channel_to_response_topic, channel_to_request_topic


class RPCClient:
    class RPCEntry:
        def __init__(self, timeout):
            self.timestamp = time.time()
            self.timeout = timeout
            self.result = None
            self.ready = threading.Event()

    def __init__(self, core: OpenModuleCore, channels=None, timeout=3., cleanup_time=300):
        if channels is None:
            channels = []
        self.core = core
        self.log = logging.getLogger("rcp-client")
        self.lock = threading.Lock()
        self.results = dict()
        self.sync_timeout = timeout
        self.cleanup_time = cleanup_time
        self.running = True

        self.channels = []
        for channel in channels:
            self.register_channel(channel)

    def register_channel(self, channel):
        if not self.running:
            raise AssertionError("Cannot register channels when rpc client is shutdown")
        if channel not in self.channels:
            self.channels.append(channel)
            topic = channel_to_response_topic(channel)
            self.log.debug("Registering channel: {}".format(topic))
            self.core.messages.register_handler(topic, RPCResponse, self.receive)

    def unregister_channel(self, channel):
        self.channels.remove(channel)
        topic = channel_to_response_topic(channel)
        self.log.debug("Unregistering channel: {}".format(topic))
        self.core.sub_socket.unsubscribe(topic)

    def cleanup_old_results(self):
        now = time.time()
        with self.lock:
            to_delete = []
            for rpc_id, entry in self.results.items():
                if now > entry.timestamp + self.cleanup_time + entry.timeout:
                    to_delete.append(rpc_id)
            for rpc_id in to_delete:
                try:
                    del self.results[rpc_id]
                except KeyError:
                    pass

    def _call(self, channel: bytes, typ: str, request: Dict, timeout: float):
        rpc_id = str(uuid4())

        request = RPCRequest(rpc_id=rpc_id, resource=self.core.config.RESOURCE, name=self.core.config.NAME,
                             request=request, type=typ)
        topic = channel_to_request_topic(channel)
        self.results[rpc_id] = self.RPCEntry(timeout=timeout)
        self.core.publish(topic=topic, message=request)
        return rpc_id

    def check_result(self, rpc_id) -> Optional[RPCResult]:
        with self.lock:
            entry = self.results.get(str(rpc_id))
            if entry and entry.result:
                self.results.pop(rpc_id, None)
                status = "ok"
                if isinstance(entry.result, dict) and entry.result.get("status"):
                    status = entry.result["status"]
                return RPCResult(status=status, response=entry.result, rpc_id=rpc_id)

    def rpc(self, channel: bytes, type: str, request: Dict, timeout: float = None, blocking=True) -> RPCResult:
        self.cleanup_old_results()

        if timeout is None:
            timeout = self.sync_timeout

        if channel not in self.channels:
            self.register_channel(channel)

        rpc_id = self._call(channel, type, request, timeout)

        if blocking:
            self.log.debug("blocking rpc, this may cause delay")
            if self.results[rpc_id].ready.wait(timeout):
                return self.check_result(rpc_id)
            else:
                raise TimeoutError
        else:
            self.log.debug("non blocking rpc sent")
            return RPCResult(status="ok", response=None, rpc_id=rpc_id)

    def shutdown(self):
        self.running = False
        for channel in self.channels:
            self.unregister_channel(channel)

    def receive(self, response: RPCResponse):
        """handler that receives, saves and cleans up rpc responses"""

        with self.lock:
            if str(response.rpc_id) in self.results:
                self.results[str(response.rpc_id)].result = response.response
                self.results[str(response.rpc_id)].ready.set()
        self.cleanup_old_results()



import re
from datetime import datetime, timedelta, tzinfo
from enum import Enum
from typing import List, Optional, Union

from dateutil import rrule, tz
from dateutil.parser import parse
from dateutil.rrule import rrulestr
from dateutil.tz import UTC
from pydantic import validator, Field

from openmodule.config import settings
from openmodule.models.base import OpenModuleModel, ZMQMessage, Gateway, timezone_validator, datetime_to_timestamp


class MediumType(str, Enum):
    lpr = "lpr"
    nfc = "nfc"
    pin = "pin"
    qr = "qr"


class AccessCategory(str, Enum):
    booked_digimon = "booked-digimon"
    booked_employee = "booked-employee"
    booked_visitor = "booked-visitor"
    permanent_digimon = "permanent-digimon"
    permanent_employee = "permanent-employee"
    filler_employee = "filler-employee"
    filler_digimon = "filler-digimon"
    filler_visitor_button = "filler-visitor-button"
    filler_visitor_unexpected = "filler-visitor-unexpected"
    unknown_category = "unknown-category"


def check_recurrence(cls, recurrence, values, **kwargs):
    if recurrence:
        if not values.get("duration"):
            raise ValueError("set a duration when using recurrence")

        try:
            if "DTSTART" not in recurrence:
                raise ValueError("recurrence must contain a DTSTART field")
            if "\n" not in recurrence:
                raise ValueError("DTSTART must be separated by a newline '\\n' character")

            rrule.rrulestr(recurrence)
        except Exception as e:
            raise ValueError(f"recurrence is not valid '{e}'") from None
        return recurrence
    else:
        return None


class Access(OpenModuleModel):
    category: AccessCategory
    start: datetime
    end: Optional[datetime] = None
    duration: Optional[int] = Field(description="duration in seconds, starting with the start time of the "
                                                "recurrence. Required if recurrence is set.")
    recurrence: Optional[str]
    zone: Optional[str]
    occupant_check: bool = False
    infos: Optional[dict] = {}
    user: str

    _check_recurrence = validator("recurrence", allow_reuse=True)(check_recurrence)
    _tz_start = timezone_validator("start")
    _tz_end = timezone_validator("end")

    _rrule_regex = re.compile(r"\s*DTSTART[:|=]+([\d\w+:]*)\s*[\n|;]*", re.IGNORECASE)

    def dict(self, **kwargs):
        data = super().dict(**kwargs)
        data["start"] = datetime_to_timestamp(data["start"])
        if data.get("end") is not None:
            data["end"] = datetime_to_timestamp(data["end"])
        return data

    def _recurrence_is_valid_at(self, native_utc_dt: datetime, timezone):
        if isinstance(timezone, str):
            timezone = tz.gettz(timezone)

        # we first construct a rrule object with dtstart as a naive localized datetime
        dtstart_string = self._rrule_regex.search(self.recurrence).group(1)
        dtstart_utc = parse(dtstart_string).replace(tzinfo=UTC)
        local_dtstart = dtstart_utc.astimezone(timezone)
        naive_dtstart = local_dtstart.replace(tzinfo=None)
        recurrence_without_dtstart = self._rrule_regex.sub("", self.recurrence)
        recurrence = rrulestr(recurrence_without_dtstart, dtstart=naive_dtstart, cache=True)

        # get the nearest event before our desired timestamp
        local_dt = native_utc_dt.replace(tzinfo=UTC).astimezone(timezone)
        naive_dt = local_dt.replace(tzinfo=None)
        naive_before = recurrence.before(naive_dt, inc=True)
        if not naive_before:
            return False

        # calculate the start/end of the recurrent event
        local_before = naive_before.replace(tzinfo=timezone)
        local_after = local_before + timedelta(seconds=self.duration)

        # in case the utc offset jumps (dst change) we always add the time difference here, because there are two
        # groups of users, and we want to work for both
        # given a dst change at 03:00 back to 02:00:
        # a) user A always arrives at 02:00, and stays 2 hours, he does not care if a dst change happened, he always
        #    stays 2 hours, regardless of the dst change, for him `start + timedelta(hours=2)` would work
        # b) user B always arrives at 04:00, so he does not care that there has been an extra hour between 2 and 4
        # oclock, he arrives at 04:00. For hom `start + timedelta(hours=2)` would miss by an hour
        utc_offset_diff = abs(local_before.utcoffset() - local_after.utcoffset())
        if utc_offset_diff.total_seconds() != 0:
            local_after += utc_offset_diff
        naive_after = local_after.replace(tzinfo=None)

        return naive_before <= naive_dt < naive_after

    def is_valid_at(self, dt: datetime, timezone: Union[str, tzinfo]):
        if dt.tzinfo:
            dt = dt.astimezone(UTC).replace(tzinfo=None)

        if self.end:
            between_start_and_end = self.start <= dt <= self.end
        else:
            between_start_and_end = self.start <= dt

        if self.recurrence:
            return between_start_and_end and self._recurrence_is_valid_at(dt, timezone)
        else:
            return between_start_and_end


class BackendRegisterRequestMessage(ZMQMessage):
    """
    sent by the controller as a request to all backends
    each backend who wants to register itself at the controller has to answer
    with a register message
    """
    type: str = "register_request"


class BackendRegisterMessage(ZMQMessage):
    """
    sent by a backend if it wants to register itself at the controller
    """
    type: str = "register"


class AccessRequest(OpenModuleModel):
    """
    The AccessRequest Model
    """
    name: str
    gateway: Optional[Gateway] = None
    medium_type: MediumType
    id: str


class MediumAccesses(OpenModuleModel):
    accesses: List[Access]
    id: str
    type: str


class AccessResponse(OpenModuleModel):
    success: bool = False
    medium: MediumAccesses


class CountMessage(ZMQMessage):
    resource: str = settings.RESOURCE
    user: str
    gateway: Gateway
    medium_type: MediumType
    id: str
    count: int
    transaction_id: str
    zone: str
    category: AccessCategory
    real: bool
    vehicle_id: Optional[str]
    infos: Optional[dict]
    access_data: Optional[dict]
    error: Optional[str]
    previous_transaction_id: Optional[List[str]]  # double_entry, choose_random error
    previous_user: Optional[str]  # user_changed error
    previous_medium_type: Optional[str]  # medium_changed, medium_id_changed error
    previous_id: Optional[str]  # medium_changed, medium_id_changed error
    chosen: Optional[dict]  # choose_random error



import re

from json.encoder import ESCAPE_ASCII

import json

from datetime import datetime
from enum import Enum

import orjson
import zmq
from dateutil.tz import UTC
from pydantic import Field, BaseModel, validator
from pydantic.main import ROOT_KEY

from openmodule import config
from openmodule.config import settings


def _donotuse(v, *, default):
    assert False, "please use json_bytes"


if config.run_checks():
    from openmodule import checks

    meta_kwargs = {"metaclass": checks.CheckingOpenModuleModel}
else:
    meta_kwargs = {}

ESCAPE_ASCII = re.compile(r'([^ -~])')


def replace(match):
    s = match.group(0)
    n = ord(s)
    if n < 0x10000:
        return '\\u%04x' % (n,)
    else:
        # surrogate pair
        n -= 0x10000
        s1 = 0xd800 | ((n >> 10) & 0x3ff)
        s2 = 0xdc00 | (n & 0x3ff)
        return '\\u%04x\\u%04x' % (s1, s2)


def _unicode_escape(data):
    return ESCAPE_ASCII.sub(replace, data)


class OpenModuleModel(BaseModel, **meta_kwargs):
    class Config:
        validate_assignment = True

        json_loads = orjson.loads
        json_dumps = _donotuse

    def json_bytes(self):
        data = self.dict()
        if self.__custom_root_type__:
            data = data[ROOT_KEY]
        return _unicode_escape(orjson.dumps(data).decode()).encode("ascii")

    def dict(self, **kwargs):
        kwargs.setdefault("by_alias", True)
        kwargs.setdefault("exclude_none", True)
        return super().dict(**kwargs)


class EmptyModel(OpenModuleModel):
    pass


def _timezone_validator(cls, dt: datetime, values, **kwargs):
    if dt and dt.tzinfo:
        return dt.astimezone(UTC).replace(tzinfo=None)
    else:
        return dt


def timezone_validator(field):
    return validator(field, allow_reuse=True)(_timezone_validator)


_EPOCH_UTC = datetime(1970, 1, 1, tzinfo=UTC)
_EPOCH_NAIVE = datetime(1970, 1, 1)


def datetime_to_timestamp(dt: datetime):
    """
    you SHOULD use this function, as it always outputs the UTC unix timestamp, regardless
    of the system's timezone. This is not an issue in production as all systems run UTC,
    but while developing it creates annoying bugs
    """
    if dt.tzinfo is None:
        return (dt - _EPOCH_NAIVE).total_seconds()
    else:
        return (dt - _EPOCH_UTC).total_seconds()


class ZMQMessage(OpenModuleModel):
    timestamp: datetime = Field(default_factory=lambda: datetime.utcnow())
    name: str
    type: str

    def __init__(self, **kwargs):
        name = kwargs.pop("name", None)
        if name is None:
            name = settings.NAME
        super().__init__(name=name, **kwargs)

    _tz_timestamp = timezone_validator("timestamp")

    def publish_on_topic(self, pub_socket: zmq.Socket, topic: bytes):
        pub_socket.send_multipart((topic, self.json_bytes()))

    def dict(self, **kwargs):
        data = super().dict(**kwargs)
        data["timestamp"] = datetime_to_timestamp(data["timestamp"])
        return data


class Direction(str, Enum):
    UNKNOWN = ""
    IN = "in"
    OUT = "out"


class Gateway(OpenModuleModel):
    gate: str = ""
    direction: Direction = ""

    def __str__(self):
        return f"{self.gate}/{self.direction}"



import logging
import threading
from collections import namedtuple
from typing import Callable, Dict, Optional, Type, Union, get_origin, get_args

import zmq
from pydantic import ValidationError, BaseModel
from pydantic.main import ROOT_KEY

from openmodule.config import settings
from openmodule.messaging import get_sub_socket, get_pub_socket, receive_message_from_socket
from openmodule.models.rpc import RPCResponse, RPCRequest
from openmodule.rpc.common import channel_to_request_topic, channel_to_response_topic
from openmodule.threading import get_thread_wrapper
from openmodule.utils.schema import Schema

CallbackEntry = namedtuple("CallbackEntry", ["timestamp", "result"])
HandlerEntry = namedtuple("HandlerEntry", ["request_class", "response_class", "handler"])


def gateway_filter(gate=None, direction=None):
    def _filter(request, message, handler):
        gateway = request.get("gateway")
        if not gateway:
            return False
        return (not gate or gate == gateway.get("gate")) and \
               (not direction or direction == gateway.get("direction"))

    return _filter


class Filter:
    def __init__(self, filter, channel, type):
        self.filter = filter
        self.channel = channel
        self.type = type

    def check(self, request, message, handler, channel):
        if (self.channel is None or channel == self.channel) and (self.type is None or self.type == message.type):
            if self.filter(request=request, message=message, handler=handler):
                return True
            return False
        # filter does not apply
        return None


class RPCServer(object):
    def __init__(self, context, config, filter_resource=True):
        self.name = config.NAME
        self.sub = get_sub_socket(context, config)
        self.pub = get_pub_socket(context, config)
        self.handlers: Dict[tuple, HandlerEntry] = {}
        self.filters = []
        self.running = True
        self.log = logging.getLogger(self.__class__.__name__)
        self.thread = None
        self.resource = None
        if filter_resource:
            if config.RESOURCE:
                self.resource = config.RESOURCE
            else:
                self.log.error(
                    "Resource filter is active, but no resource is available. All RPCs requests will be processed. "
                    "This is ok for debugging / testing containers, but must not happen in production."
                )

    def add_filter(self, filter: Callable[[], bool], channel=None, type=None):
        """
        :param filter: filter function with paramters reqeust, message, handler
        :param channel: specific channel to filter or all channels if None
        :param type: specific type to filter or all types if None
        """
        self.filters.append(Filter(filter=filter, channel=channel, type=type))

    def shutdown(self, timeout=3):
        self.running = False
        if self.thread:
            self.thread.join(timeout=timeout)

    def register_handler(self, channel: str, type: str, request_class: Type[BaseModel], response_class: Type[BaseModel],
                         handler: Callable, register_schema=True):
        """
        :param channel: rpc channel you want to subscribe to, patterns/wildcards are not supported
        :param type: the request type
        :param request_class: Class definition of the request
        :param response_class: Class definition of the response
        :param handler: request handler of the form (request:dict, meta:dict) -> any
                        if the handler returns a dict which contains the key "status",
                        then the value of "status" will be used as the rpc response status.
                        There are no other reserved keys except "status".
        :param register_schema: register the schema and models of this rpc during testing to create an automatic doc
        """
        if (channel, type) in self.handlers:
            raise ValueError(f"handler for {channel}:{type} is already registered")

        self.log.debug("register handler {}:{} -> {}".format(channel, type, handler))
        self.sub.subscribe(channel_to_request_topic(channel.encode("ascii")))
        self.handlers[(channel, type)] = HandlerEntry(request_class, response_class, handler)
        if register_schema:
            Schema.save_rpc(channel, type, request_class, response_class, handler)

    def _channel_from_topic(self, topic: bytes) -> str:
        return topic.split(b"-", 2)[-1].decode("ascii")

    def should_process_message(self, request, message, handler, channel):
        if self.filters:
            result = True
            for filter in self.filters:
                ok = filter.check(request=request, message=message, handler=handler, channel=channel)
                if ok:
                    return True
                # filter applies and is not ok
                elif ok is False:
                    result = False
            return result
        return True

    def find_handler(self, channel: str, type: str) -> HandlerEntry:
        return self.handlers.get((channel, type))

    def run_as_thread(self):
        assert not self.thread, "cannot run the same rpc server multiple times as thread"
        self.thread = threading.Thread(target=get_thread_wrapper(self.run))
        self.thread.start()
        return self.thread

    def _response_to_dict(self, response):
        data = response.dict()
        if response.__custom_root_type__:
            data = data[ROOT_KEY]
        return data

    def process_rpc(self, channel, message: RPCRequest) -> Optional[Dict]:
        if self.resource and message.resource and self.resource != message.resource:
            self.log.debug("a rpc was not processed because the resource did not mach. "
                           "my resource: %s, resource specified in rpc: %s", self.resource, message.resource)
            return None

        if settings.LOG_LEVEL == logging.DEBUG:
            self.log.debug("received RPC channel: %s, type: %s, request: %s", channel, message.type, message.request)
        else:
            self.log.info("received RPC channel: %s, type: %s", channel, message.type)

        handler: HandlerEntry = self.find_handler(channel, message.type)
        if handler:
            try:
                request = handler.request_class(**message.request)
            except ValidationError as e:
                return {"status": "validation_error", "exception": e.json()}

            try:
                if not self.should_process_message(request, message, handler, channel):
                    return None
            except Exception as e:
                self.log.exception("exception in message filter. request not processed and error returned")
                return {"status": "filter_error", "exception": str(e)}

            try:
                response = handler.handler(request, message)
                if isinstance(response, handler.response_class):
                    # shortcut prevent serialization and de-serialization if the correct response
                    # is already returned
                    return self._response_to_dict(response)
                else:
                    response = handler.response_class.parse_obj(response or {})
                    return self._response_to_dict(response)
            except Exception as e:
                self.log.exception("exception in handler {}:{}".format(channel, message.type))
                return {"status": "handler_error", "exception": str(e)}
        else:
            self.log.warning("no handler found for {}:{}".format(channel, message.type))
            return None

    def run(self):
        poller = zmq.Poller()
        poller.register(self.sub, zmq.POLLIN)
        try:
            while self.running:
                socks = dict(poller.poll(timeout=1000))
                if socks.get(self.sub) != zmq.POLLIN:
                    continue
                topic, message = receive_message_from_socket(self.sub)
                if topic is None:
                    continue
                channel = self._channel_from_topic(topic)
                try:
                    message = RPCRequest(**message)
                    message_type = message.type
                    rpc_id = message.rpc_id
                except ValidationError as e:
                    message_type = "unknown"
                    rpc_id = None
                    response = {"status": "validation_error", "exception": e.json()}
                else:
                    response = self.process_rpc(channel, message)

                if response is not None:
                    self.log.debug("response: {}".format(response))
                    if "status" not in response:
                        response["status"] = "ok"

                    result = RPCResponse(
                        type=message_type,
                        rpc_id=rpc_id,
                        response=response,
                        name=self.name
                    )
                    result.publish_on_topic(
                        self.pub,
                        channel_to_response_topic(channel.encode("ascii")),
                    )
        except zmq.ContextTerminated:
            pass
        finally:
            self.sub.close()
            self.pub.close()



import time
from _weakrefset import WeakSet

from openmodule.dispatcher import MessageDispatcher
from openmodule.models.base import Direction, Gateway, OpenModuleModel
from openmodule_test.utils import DeveloperError


class TestGateAccessMessage(OpenModuleModel):
    gateway: Gateway


_all_gates = WeakSet()


class TestGate:
    __test__ = False

    _wait_sleep = 0.05
    _wait_called_since_reset = False

    open_count = 0
    reject_count = 0

    def __init__(self, gate: str, direction: Direction, dispatcher: MessageDispatcher):
        self.gate = gate
        self.direction = direction
        dispatcher.register_handler(b"access_accept", TestGateAccessMessage, self._access_accept, register_schema=False)
        dispatcher.register_handler(b"access_reject", TestGateAccessMessage, self._access_reject, register_schema=False)
        _all_gates.add(self)

    @classmethod
    def reset_all_counts(cls):
        for x in _all_gates:
            x.reset_call_count()

    def reset_call_count(self):
        self.open_count = 0
        self.reject_count = 0
        self._wait_called_since_reset = False

    def _access_accept(self, message: TestGateAccessMessage):
        if message.gateway.gate == self.gate:
            self.open_count += 1

    def _access_reject(self, message: TestGateAccessMessage):
        if message.gateway.gate == self.gate:
            self.reject_count += 1

    def assert_opened(self):
        assert self.open_count > 0

    def assert_rejected(self):
        assert self.open_count == 0 and self.reject_count > 0

    def wait_for_open(self, timeout: float = 3, minimum_open_count=1):
        self._wait_for("open_count", timeout=timeout, minimum_count=minimum_open_count)

    def wait_for_reject(self, timeout: float = 3, minimum_open_count=1):
        self._wait_for("reject_count", timeout=timeout, minimum_count=minimum_open_count)
        try:
            # we always wait half as long for any open messages which may arrive late, so the sequence
            # Reject -> Accept, does correctly raise a TimeoutError
            self._wait_called_since_reset = False
            self.wait_for_open(timeout=timeout / 2)
        except TimeoutError:
            pass
        self.assert_rejected()

    def _wait_for(self, var_name: str, timeout: float = 3, minimum_count=1):
        if self._wait_called_since_reset:
            # Explanation? -> see MockEvent.wait_for_call
            raise DeveloperError(
                "The test-case MUST reset the call count to zero on its own before calling wait_for_call.\n"
                "This is in order to prevent a race-condition which is impossible to prevent otherwise.\n"
                "Please call `.reset_counts()` or `TestGate.reset_all_counts()`"
            )
        self._wait_called_since_reset = True
        for x in range(int(timeout // self._wait_sleep)):
            time.sleep(self._wait_sleep)
            if getattr(self, var_name) >= minimum_count:
                return
        if getattr(self, var_name) < minimum_count:
            raise TimeoutError()



import contextlib
import logging
import random
import string
import time
from typing import Optional, Tuple, Dict

import orjson
import zmq

from openmodule.dispatcher import MessageDispatcher
from openmodule.models.base import ZMQMessage


def get_sub_socket(context, config, linger=0) -> zmq.Socket:
    socket = context.socket(zmq.SUB)
    socket.setsockopt(zmq.LINGER, linger)
    socket.connect(config.BROKER_PUB)
    return socket


def get_pub_socket(context, config, linger=100) -> zmq.Socket:
    socket = context.socket(zmq.PUB)
    socket.setsockopt(zmq.LINGER, linger)
    socket.connect(config.BROKER_SUB)
    return socket


def wait_for_connection(dispatcher: MessageDispatcher, pub_socket, pub_lock=None, timeout=100):
    """
    waits until the pub_socket's messages can be received at the dispatcher.
    This dispatcher needs to run in a seoarate thread as this method is blocking
    :param pub_lock: optionally locks the lock during publishing
    """

    pub_lock = pub_lock or contextlib.nullcontext()
    received = False

    def handler(_):
        nonlocal received
        received = True

    # topic needs to start with __ping because these messages are ignored by the deeplog
    random_topic = "__ping_" + "".join(random.choices(string.ascii_letters, k=10))
    random_topic = random_topic.encode()

    listener = dispatcher.register_handler(random_topic, ZMQMessage, handler, register_schema=False)

    check_delay = 0.1
    check_iterations = int(timeout / check_delay) + 1

    message = ZMQMessage(type="connection-check")
    for x in range(check_iterations):
        with pub_lock:
            message.publish_on_topic(pub_socket, random_topic)

        time.sleep(check_delay)
        if received:
            break

    dispatcher.unregister_handler(listener)
    if not received:
        raise TimeoutError("could not connect to the message broker in time")


def receive_message_from_socket(sub_socket: zmq.Socket) -> Tuple[Optional[bytes], Optional[Dict]]:
    """
    :param sub_socket: zmq socket to receive
    :return: returns None if an invalid message was received
             otherwise returns a tuple containing (topic: bytes, message: dict)
    """
    parts = sub_socket.recv_multipart()
    if len(parts) != 2:
        logging.error(f"received a zmq message with an invalid number of parts: {len(parts)}")
        return None, None

    topic, message = parts

    try:
        message_dict = orjson.loads(message)
    except orjson.JSONDecodeError as e:
        logging.error(f"received a zmq message with an invalid json. parsing error: {str(e)}")
        return None, None
    else:
        if not isinstance(message_dict, dict):
            logging.error(f"received a zmq message which was not a dict but a {type(message)}")
            return None, None
        else:
            return topic, message_dict



import functools
import json
import os
from textwrap import dedent
from typing import Callable, Any


def active(f):
    def wrapper(*args, **kwargs):
        if os.environ.get("TESTING") == "True":
            f(*args, **kwargs)
        return None

    return wrapper


class RPCSchema:
    def __init__(self, channel, typ, request, response, handler: Callable[[Any], Any]):
        self.channel = channel if isinstance(channel, str) else channel.decode()
        self.type = str(typ)
        self.request = request.__name__
        self.response = response.__name__
        self.description = dedent(handler.__doc__.strip()) if handler.__doc__ else ""

    def render(self):
        return dict(post=dict(summary=self.description, tags=["RPCs"],
                              requestBody=dict(description=self.request, content={
                                  "application/json": dict(schema={"$ref": f"#/components/schemas/{self.request}"})}),
                              responses={"200": dict(description=self.response, content={
                                  "application/json": dict(
                                      schema={"$ref": f"#/components/schemas/{self.response}"})})}))

    @property
    def key(self):
        return f"/{self.channel}/{self.type}"


class MessageSchema:
    def __init__(self, topic, message_class, handler: Callable[[Any], Any], filter):
        self.topic = topic if isinstance(topic, str) else topic.decode()
        self.message_class = message_class.__name__
        self.filter = filter
        self.handler = handler.__name__
        self.description = dedent(handler.__doc__.strip()) if handler.__doc__ else ""

    def render(self):
        res = dict(put=dict(summary=self.description, tags=["Messages"],
                            requestBody=dict(description=self.message_class, content={
                                "application/json": dict(
                                    schema={"$ref": f"#/components/schemas/{self.message_class}"})})))
        res["put"].update(self.render_filter())
        return res

    def render_filter(self):
        if not self.filter:
            return {}
        res = []
        for key, value in self.filter.items():
            res.append({"in": "filter", "name": key, "value": value})
        return dict(parameters=res)

    @property
    def key(self):
        return f"/{self.topic}/{self.handler}"

    def check_new(self, message_list):
        for message in message_list:
            if self.topic == message.topic and self.handler == message.handler and self.filter == message.filter:
                return False
        return True


class SchemaClass:
    """
    Renders a valid OpenApi JSON Schema of rpcs and pydantic models that can be displayed  with Swagger
    """
    rpcs: dict
    models: set
    messages: dict

    def __init__(self):
        self.rpcs = dict()
        self.models = set()
        self.messages = dict()

    @classmethod
    def get_correct_handler(cls, handler, validate) -> Callable[[Any], Any]:
        if isinstance(handler, functools.partial):
            func = handler.func
        else:
            func = handler

        if validate:
            assert func.__doc__, f"You need to describe the message handler {func} with a doc string"
        return func

    @active
    def save_rpc(self, channel, typ, request, response, handler, validate=True):
        self.save_model(request)
        self.save_model(response)

        func = self.get_correct_handler(handler, validate)
        if func:
            schema = RPCSchema(channel, typ, request, response, func)
            self.rpcs[schema.key] = schema

    @active
    def save_model(self, model):
        self.models.add(model)

    @active
    def save_message(self, topic, message_class, handler, filter=None, validate=True):
        self.save_model(message_class)
        func = self.get_correct_handler(handler, validate)
        if func:
            schema = MessageSchema(topic, message_class, func, filter)
            if self.messages.get(schema.key):
                if schema.check_new(self.messages[schema.key]):
                    self.messages[schema.key].append(schema)
            else:
                self.messages[schema.key] = [schema]

    def __str__(self):
        return json.dumps(self.render_schema())

    def clear(self):
        self.rpcs.clear()
        self.models.clear()
        self.messages.clear()

    def render_schema(self):
        res = dict(openapi="3.0.0", info=dict(title="RPC & Message API", version="1.0.0"))
        paths = self.render_rpcs()
        paths.update(self.render_messages())
        if paths:
            res["paths"] = paths
        models = self.render_models()
        if models:
            res["components"] = dict(schemas=models)
        return res

    def render_rpcs(self):
        result = dict()
        for key, value in self.rpcs.items():
            result[key] = value.render()
        return result

    def render_messages(self):
        result = dict()
        for key, messages in self.messages.items():
            if len(messages) > 1:
                for cnt, message in enumerate(messages):
                    result[f"{key} - {cnt + 1}"] = message.render()
            else:
                result[key] = messages[0].render()
        return result

    def render_models(self):
        result = dict()
        for model in self.models:
            model_schema = model.schema().copy()
            definitions = model_schema.pop("definitions", {})
            result[model.__name__] = model_schema
            for key, value in definitions.items():
                result[key] = value
        result = json.loads(json.dumps(result).replace("#/definitions/", "#/components/schemas/"))
        return result

    @active
    def to_file(self, file_name=None):
        if self.models:
            file_name = file_name or os.environ.get("SCHEMA_FILE", "../schemas.json")
            if os.path.exists(file_name):
                with open(file_name, "r") as file:
                    schema = json.loads(file.read())
                    new_schema = self.render_schema()
                    if schema.get("paths"):
                        schema["paths"].update(new_schema.get("paths", {}))
                    else:
                        schema["paths"] = new_schema.get("paths")
                    if schema.get("components"):
                        schema["components"]["schemas"].update(new_schema.get("components", {}).get("schemas", {}))
                    else:
                        schema["components"] = dict(schemas=new_schema.get("components", {}).get("schemas", {}))

                    schema["components"]["schemas"] = {k: v for k, v in sorted(schema["components"]["schemas"].items(),
                                                                               key=lambda item: item[0])}
                    data = json.dumps(schema)
            else:
                data = str(self)
            with open(file_name, "w") as file:
                file.write(data)


Schema = SchemaClass()



from unittest import TestCase

import json
from textwrap import dedent

from openmodule.config import override_settings
from openmodule.models.backend import CountMessage, AccessRequest, Access, AccessCategory, check_recurrence
from openmodule.models.base import Gateway
from openmodule.rpc.server import RPCServer
from openmodule.utils.backend import Backend
from openmodule_test.backend import TestBackend, BackendTestMixin
from openmodule_test.eventlistener import MockEvent
from openmodule_test.rpc import RPCServerTestMixin


class ModelTest(TestCase):
    def test_count_message(self):
        messages = dedent("""  [{  "real": true,
                    "category": "permanent-employee",
                    "medium_type": "lpr",
                    "name": "count-test",
                    "zone": "zone1",
                    "timestamp": 1536154434.152993,
                    "count": 1,
                    "gateway": {
                        "gate": "gate1",
                        "direction": "in"
                    },
                    "user": "userprefix-asdf_GASDF1\u00dcSER",
                    "access_data": {
                        "on_presence_loop": true,
                        "has_presence_loop": false,
                        "decision": "accepted"
                    },
                    "type": "count",
                    "id": "GASD\u00c41",
                    "transaction_id": "dc4e4af3-5ff4-4258-a310-089701ef37e8" }, 
                    { "real": false,
                    "category": "permanent-employee",
                    "medium_type": "lpr",
                    "name": "count-test",
                    "zone": "zone1",
                    "timestamp": 1536154434.19146,
                    "count": 0,
                    "gateway": {
                        "gate": "automatically_deleted",
                        "direction": "out"
                    },
                    "user": "userprefix-asdf_GASDF1\u00dcSER",
                    "type": "count",
                    "id": "GASD\u00c41",
                    "transaction_id": "dc4e4af3-5ff4-4258-a310-089701ef37e8"},
                     {  "real": true,
                    "category": "permanent-employee",
                    "medium_type": "lpr",
                    "name": "count-test",
                    "zone": "zone1",
                    "previous_transaction_id": ["dc4e4af3-5ff4-4258-a310-089701ef37e8"],
                    "timestamp": 1536154434.192287,
                    "count": 1,
                    "gateway": {
                        "gate": "gate1",
                        "direction": "in"
                    },
                    "user": "userprefix-asdf_GASDF1\u00dcSER",
                    "error": "double_entry",
                    "access_data": {
                        "on_presence_loop": true,
                        "has_presence_loop": false,
                        "decision": "accepted"
                    },
                    "type": "count",
                    "id": "GASD\u00c41",
                    "transaction_id": "0337597c-f125-44e0-b38f-a6537e7d38d1" }]
                """).strip()
        messages = json.loads(messages)
        for message in messages:
            model = CountMessage(**message)
            self.assertTrue(bool(model.transaction_id))

    def test_access_recurrence(self):
        kwargs = dict(user="asdf", category="permanent-employee", start=3,
                      recurrence="DTSTART:20180702T123000\nFREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR")
        with self.assertRaises(ValueError) as e:
            Access(**kwargs)
        self.assertIn("set a duration", str(e.exception))

        kwargs["duration"] = 0
        with self.assertRaises(ValueError) as e:
            Access(**kwargs)
        self.assertIn("set a duration", str(e.exception))

        kwargs["duration"] = 1000
        access = Access(**kwargs)
        self.assertEqual(1000, access.duration)

        kwargs["recurrence"] = "asdf"
        with self.assertRaises(ValueError) as e:
            Access(**kwargs)
        self.assertIn("recurrence is not valid", str(e.exception))

    def test_check_recurrence_validator(self):
        cls = None

        # due to too much legacy data which contains "", we allow empty string, but it is converted to none internally
        check_recurrence(cls, "", {})
        access = Access(start=0, user="test", category=AccessCategory.permanent_digimon, recurrence="")
        self.assertIsNone(access.recurrence)

        with self.assertRaises(ValueError) as e:
            check_recurrence(cls, "DTSTART\n", {"duration": "100"})
        self.assertIn("recurrence is not valid", str(e.exception))

        with self.assertRaises(ValueError) as e:
            check_recurrence(cls, "\nRRULE:FREQ=WEEKLY;INTERVAL=1;WKST=MO", {"duration": 3600})
        self.assertIn("DTSTART", str(e.exception))

        with self.assertRaises(ValueError) as e:
            check_recurrence(cls, "DTSTART:20210304T161900Z;RRULE:FREQ=WEEKLY;INTERVAL=1;WKST=MO", {"duration": 3600})
        self.assertIn("separated by a newline", str(e.exception))

        check_recurrence(cls, None, {})
        check_recurrence(cls, "DTSTART:20210304T161900Z\nRRULE:FREQ=WEEKLY;INTERVAL=1;WKST=MO", {"duration": 3600})

    def test_category(self):
        with self.assertRaises(Exception) as e:
            AccessCategory("test_exception")
        self.assertIn("is not a valid accesscategory", str(e.exception).lower())

        with self.assertRaises(Exception) as e:
            Access(user="asdf", category="permanent_employee", start=3,
                   recurrence="DTSTART:20180702T123000\nFREQ=WEEKLY;BYDAY=MO,TU,WE,TH,FR")
        self.assertIn("not a valid enumeration", str(e.exception))


@override_settings(USER_PREFIX="backend")
class NotImplementedTestCase(BackendTestMixin, TestCase):
    backend_class = Backend

    # noinspection PyTypeChecker
    def test_not_implemented(self):
        with self.assertRaises(NotImplementedError):
            self.backend.check_in(None)
        with self.assertRaises(NotImplementedError):
            self.backend.check_out(None)
        with self.assertRaises(NotImplementedError):
            self.backend.check_access(None)

@override_settings(USER_PREFIX="backend")
class FunctionTest(BackendTestMixin, TestCase):
    backend_class = TestBackend

    def test_check_in(self):
        transaction_id = self.check_in()
        message = self.messages_by_transaction.get(transaction_id)
        self.assertEqual(None, message.error)

        transaction_id = self.check_in_double_entry(transaction_id)
        message = self.messages_by_transaction.get(transaction_id)
        self.assertEqual("double_entry", message.error)
        self.assertNotEqual(None, message.previous_transaction_id)

        with self.assertRaises(Exception) as e:
            self.check_in(gate="error")
        self.assertIn("test_exception", str(e.exception))

    def check_in_out(self):
        transaction_id = self.check_in()
        self.assertNotEqual(None, self.messages_by_transaction.get(transaction_id))

        self.check_out(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))

    def test_check_out(self):
        with self.assertRaises(Exception) as e:
            self.check_out("asdf")
        self.assertIn("asdf", str(e.exception))

        transaction_id = self.check_in(send=False)
        self.messages_by_transaction[transaction_id].gateway.gate = "error"
        with self.assertRaises(Exception) as e:
            self.check_out(transaction_id)
        self.assertIn("test_exception", str(e.exception))

        transaction_id = self.check_in(send=False)
        message = self.check_out(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual(None, message.error)

        # category_changed
        transaction_id = self.check_in(send=False)
        message = self.check_out_category_changed(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual("category_changed", message.error)

        transaction_id = self.check_in(send=False)
        with self.assertRaises(Exception) as e:
            self.check_out_category_changed(transaction_id, category="permanent-employee")
        self.assertIn("category needs to change", str(e.exception))

        # medium_changed
        transaction_id = self.check_in(send=False)
        message = self.check_out_medium_changed(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual("medium_type_changed", message.error)
        self.assertNotEqual(None, message.previous_id)
        self.assertNotEqual(None, message.previous_medium_type)

        transaction_id = self.check_in(send=False)
        with self.assertRaises(Exception) as e:
            self.check_out_medium_changed(transaction_id, medium="lpr")
        self.assertIn("medium needs to change", str(e.exception))

        # medium_id_changed
        transaction_id = self.check_in(send=False)
        message = self.check_out_medium_id_changed(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual("medium_id_changed", message.error)
        self.assertNotEqual(None, message.previous_id)

        transaction_id = self.check_in(send=False)
        medium_id = self.messages_by_transaction[transaction_id].id
        with self.assertRaises(Exception) as e:
            self.check_out_medium_id_changed(transaction_id, medium_id=medium_id)
        self.assertIn("medium_id needs to change", str(e.exception))

        # user_changed
        transaction_id = self.check_in(send=False)
        message = self.check_out_user_changed(transaction_id)
        self.assertEqual(None, self.messages_by_transaction.get(transaction_id))
        self.assertEqual("user_changed", message.error)
        self.assertNotEqual(None, message.previous_user)

        transaction_id = self.check_in(send=False)
        with self.assertRaises(Exception) as e:
            self.check_out_user_changed(transaction_id, user="api-backend_user")
        self.assertIn("user needs to change", str(e.exception))

    def test_access(self):
        result = self.check_auth()
        self.assertEqual(True, result.success)
        self.assertNotEqual([], result.medium.accesses)

        result = self.check_auth(gate="empty")
        self.assertEqual(True, result.success)
        self.assertEqual([], result.medium.accesses)

        result = self.check_auth(gate="error")
        self.assertEqual(False, result.success)


@override_settings(USER_PREFIX="backend")
class RpcTest(BackendTestMixin, RPCServerTestMixin, TestCase):
    rpc_channels = ["backend"]
    topics = [b"backend", b"count", b"healthz"]
    backend_class = TestBackend

    def setUp(self):
        super().setUp()
        self.server = RPCServer(config=self.zmq_config(), context=self.zmq_context())
        self.server.run_as_thread()
        self.backend.register_rpcs(self.server)
        self.wait_for_rpc_server(self.server)

    def tearDown(self):
        self.server.shutdown()
        super().tearDown()

    def test_access_rpc(self):
        # only test if rpc is passed on to function, function is tested in previous TestCase
        gateway = Gateway(gate="gate", direction="in")
        request = AccessRequest(gateway=gateway, name=self.core.config.NAME, medium_type="lpr", id="medium_id")
        response = self.rpc("backend", "auth", request=request.dict())
        self.assertRPCSuccess(response)
        self.assertIn(request.id, self.backend.accessed)

    def send_count_message(self, gate="gate", direction="in"):
        transaction_id = self.create_count_message(direction=direction, gate=gate)
        self.zmq_client.send("count", **self.messages_by_transaction[transaction_id].dict())

    def test_registration(self):
        # auto register on start
        message = self.zmq_client.wait_for_message_on_topic(topic=b"backend", timeout=5)
        self.assertEqual("register", message["type"])

        self.zmq_client.send("backend", type="register_request")
        message = self.zmq_client.wait_for_message_on_topic(topic=b"backend", timeout=5)
        self.assertEqual("register_request", message["type"])

        message = self.zmq_client.wait_for_message_on_topic(topic=b"backend", timeout=5)
        self.assertEqual("register", message["type"])

    def test_check_in(self):
        self.send_count_message()
        self.backend.wait_for_message_process()
        self.assertEqual(1, self.backend.checked_in)

        MockEvent.reset_all_call_counts()
        with self.assertLogs() as cm:
            self.send_count_message(gate="error")
            self.backend.wait_for_message_process()
        self.assertEqual(1, self.backend.checked_in)
        self.assertIn("error in check_in", cm.output[0])

    def test_check_out(self):
        self.send_count_message()
        self.backend.wait_for_message_process()
        self.assertEqual(1, self.backend.checked_in)

        MockEvent.reset_all_call_counts()
        with self.assertLogs() as cm:
            self.send_count_message(gate="error", direction="out")
            self.backend.wait_for_message_process()
        self.assertEqual(1, self.backend.checked_in)
        self.assertIn("error in check_out", cm.output[0])

        MockEvent.reset_all_call_counts()
        self.send_count_message(direction="out")
        self.backend.wait_for_message_process()
        self.assertEqual(0, self.backend.checked_in)

    def test_datetime_is_converted_to_unix_timestamp(self):
        gateway = Gateway(gate="entry", direction="in")
        request = AccessRequest(gateway=gateway, name=self.core.config.NAME, medium_type="lpr", id="GTEST1")
        response = self.rpc("backend", "auth", request=request.dict())
        medium = response["response"]["medium"]
        self.assertTrue(isinstance(medium["accesses"][0]["start"], (int, float)))
        self.assertTrue(isinstance(medium["accesses"][0]["end"], (int, float)))



from sqlalchemy.orm import Session


def update_query(db: Session, query, values: dict) -> int:
    """
    in order to update via a query in sqlite we have to no synchronize the session
    after this operation all db objects are expired (same as if you access the object
    after the transaction)

    :return: number of elements updated
    """
    res = query.update(values, synchronize_session=False)
    db.expire_all()
    return res


def delete_query(db: Session, query) -> int:
    """
    in order to delete via a query in sqlite we have to not synchronize the session
    after this operation all db objects are expired (same as if you access the object
    after the transaction)

    :return: number of elements deleted
    """
    res = query.delete(synchronize_session=False)
    db.expire_all()
    return res



import os
import sys
import threading
import types
import warnings
from typing import Optional

from alembic import command, context
from alembic.config import Config
from sqlalchemy import create_engine, event, MetaData
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker, Session
from sqlalchemy.pool import StaticPool

from openmodule.checks import check_invalid_database_column_type


@event.listens_for(Engine, "connect")
def set_sqlite_pragma(dbapi_connection, connection_record):
    cursor = dbapi_connection.cursor()
    cursor.execute("PRAGMA foreign_keys=ON")
    cursor.close()


def alembic_config(connection, alembic_path):
    alembic_cfg = Config(os.path.join(alembic_path, "alembic.ini"),
                         attributes={
                             "configure_logging": False,
                             "connection": connection,
                         })
    alembic_cfg.set_main_option("script_location", os.path.join(alembic_path, "alembic"))
    return alembic_cfg


def migrate_database(connection, alembic_path=None):
    if alembic_path is None:
        alembic_path = os.path.join(os.getcwd(), "database")
    assert os.path.exists(os.path.abspath(alembic_path)), f"alembic path {os.path.abspath(alembic_path)} does not exist"
    config = alembic_config(connection, alembic_path)
    command.upgrade(config, "head")


active_databases = {}


def register_bases(bases, show_deprecation_warning=True):
    if show_deprecation_warning:
        warnings.warn(
            '\n\n`register_bases([...])` followed by `from openmodule.database.env import *` is deprecated.\n '
            'Please replace these lines with `run_env_py([bases...])`\n',
            DeprecationWarning)

    target_metadata = MetaData()

    if not isinstance(bases, list):
        bases = [bases]

    for base in bases:
        for table in base.metadata.tables.values():
            for x in table.columns:
                check_invalid_database_column_type(x.type)
            table.tometadata(target_metadata)
    context.config.attributes["target_metadata"] = target_metadata


def run_env_py(bases):
    register_bases(bases, show_deprecation_warning=False)
    # noinspection PyUnresolvedReferences
    import openmodule.database.env
    del sys.modules["openmodule.database.env"]


def database_path(db_folder, db_name):
    return os.path.join(db_folder, db_name) + ".sqlite3"


def get_database(db_folder: str, name: str, alembic_path=None):
    global active_databases
    tmp = database_path(db_folder, name)
    assert active_databases.get(tmp) is None, f"database {tmp} already exists," \
                                              f" check if it was shutdown before a new one was created"
    os.makedirs(db_folder, exist_ok=True)
    path = f"sqlite:///{tmp}"
    engine = create_engine(path, poolclass=StaticPool, connect_args={'check_same_thread': False})
    migrate_database(engine, alembic_path)
    active_databases[tmp] = engine

    return engine


class DatabaseContext:
    def __init__(self, database: 'Database', expire_on_commit=True):
        self.database = database
        self.expire_on_commit = expire_on_commit

    def __enter__(self) -> Session:
        return self.database.__enter__(expire_on_commit=self.expire_on_commit)

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.database.__exit__(exc_type, exc_val, exc_tb)


class SessionWrapper(Session):
    _closed = False

    def __getattribute__(self, item):
        attr = super().__getattribute__(item)
        if isinstance(attr, types.MethodType) and self._closed:
            raise AssertionError("Session is already closed")
        return attr

    def close(self):
        super().close()
        self._closed = True


class Database:
    active_session: Optional[Session]

    def __init__(self, database_folder, name="database", alembic_path=None):
        self.db_folder = database_folder
        self.name = name
        self._engine = get_database(database_folder, name, alembic_path)
        self._session = sessionmaker(bind=self._engine, class_=SessionWrapper)
        self.active_session = None
        self.lock = threading.RLock()

    def is_open(self):
        return bool(self._session)

    def shutdown(self):
        assert self.is_open(), "database is already closed, you called shutdown twice somewhere"

        with self.lock:
            self._session = None
            global active_databases
            active_databases.pop(database_path(self.db_folder, self.name), None)

    def __call__(self, expire_on_commit=True) -> DatabaseContext:
        return DatabaseContext(self, expire_on_commit=expire_on_commit)

    def __enter__(self, expire_on_commit=True) -> Session:
        assert self._session, "Session is already closed"
        self.lock.acquire()
        self.active_session = self._session(expire_on_commit=expire_on_commit)
        return self.active_session

    def flush(self, objects=None):
        assert self._session and self.active_session, "Session is already closed"
        self.active_session.flush(objects)

    def __exit__(self, exc_type, exc_val, exc_tb):
        try:
            if exc_type is None:
                self.active_session.commit()
            else:
                self.active_session.rollback()
        finally:
            self.active_session.close()
            self.active_session = None
            self.lock.release()



import json
import os
from functools import partial
from unittest import TestCase

from openmodule.config import settings
from openmodule.models.backend import AccessRequest, AccessResponse
from openmodule.rpc.server import RPCServer
from openmodule.utils.schema import Schema
from openmodule_test.core import OpenModuleCoreTestMixin
from openmodule_test.rpc import RPCServerTestMixin

file_name = "../schemas.json"

def test(*args, **kwargs):
    """
    Test
    """
    return None


def test2(*args, **kwargs):
    """
    Test
    """
    return None


def test1(*args, **kwargs):
    return None


def test3(*args, **kwargs):
    return None


class SchemaTest(TestCase):
    def tearDown(self) -> None:
        super().setUp()
        try:
            os.remove(file_name)
        except:
            pass
        Schema.clear()
        super().tearDown()

    def setUp(self) -> None:
        super().setUp()
        os.environ["TESTING"] = "True"
        try:
            os.remove(file_name)
        except:
            pass
        Schema.clear()

    def test_models(self):
        Schema.save_model(AccessRequest)
        schema = Schema.render_schema()
        length = len(schema["components"]["schemas"])
        self.assertEqual(1, len(Schema.models))
        self.assertIn("AccessRequest", schema["components"]["schemas"])
        self.assertEqual(None, schema.get("paths"))

        # same model
        Schema.save_model(AccessRequest)
        schema = Schema.render_schema()
        self.assertEqual(1, len(Schema.models))
        self.assertIn("AccessRequest", schema["components"]["schemas"])
        self.assertEqual(None, schema.get("paths"))
        self.assertEqual(length, len(schema["components"]["schemas"]))

        Schema.save_model(AccessResponse)
        schema1 = Schema.render_schema()
        self.assertEqual(2, len(Schema.models))
        self.assertIn("AccessRequest", schema["components"]["schemas"])
        self.assertIn("AccessResponse", schema1["components"]["schemas"])
        self.assertGreaterEqual(len(schema1["components"]["schemas"]), length)

    def test_rpcs(self):
        Schema.save_rpc("backend", "auth", AccessRequest, AccessResponse, test)
        self.assertEqual(1, len(Schema.rpcs))

        schema = Schema.render_schema()
        self.assertIn("/backend/auth", schema["paths"])
        self.assertEqual(len(schema["paths"]), 1)

        Schema.save_rpc("backend", "auth", AccessRequest, AccessResponse, test)
        schema = Schema.render_schema()
        self.assertIn("/backend/auth", schema["paths"])
        self.assertEqual(1, len(Schema.rpcs))
        self.assertEqual(len(schema["paths"]), 1)

        Schema.save_rpc("backend", "abc", AccessRequest, AccessResponse, test)
        schema = Schema.render_schema()
        self.assertIn("/backend/abc", schema["paths"])
        self.assertIn("/backend/auth", schema["paths"])
        self.assertEqual(2, len(Schema.rpcs))
        self.assertEqual(len(schema["paths"]), 2)

    def test_messages(self):
        Schema.save_message("message_topic", AccessRequest, test)
        self.assertEqual(1, len(Schema.messages))
        schema = Schema.render_schema()
        self.assertIn("/message_topic/test", schema["paths"])
        self.assertEqual(len(schema["paths"]), 1)

        Schema.save_message("message_topic", AccessRequest, test)
        self.assertEqual(1, len(Schema.messages))
        schema = Schema.render_schema()
        self.assertIn("/message_topic/test", schema["paths"])
        self.assertEqual(len(schema["paths"]), 1)

        with self.assertRaises(AssertionError):
            Schema.save_message("message_topic", AccessRequest, test1)

        Schema.save_message("message_topic", AccessRequest, test2)
        schema = Schema.render_schema()
        self.assertIn("/message_topic/test", schema["paths"])
        self.assertIn("/message_topic/test2", schema["paths"])
        self.assertEqual(2, len(schema["paths"]))
        self.assertEqual(2, len(Schema.messages))

    def test_message_filters(self):
        Schema.save_message("message_topic", AccessRequest, test, {"a": 1})
        Schema.save_message("message_topic", AccessRequest, test, {"a": 2})
        self.assertEqual(1, len(Schema.messages))
        self.assertEqual(2, len(Schema.messages["/message_topic/test"]))

        schema = Schema.render_messages()
        self.assertIn("/message_topic/test - 2", schema)
        self.assertIn("/message_topic/test - 2", schema)

    def test_file(self):
        def get_file_data():
            with open(file_name, "r") as file:
                return json.loads(file.read())

        Schema.to_file(file_name)
        with self.assertRaises(FileNotFoundError):
            data = get_file_data()

        Schema.save_rpc("backend", "auth", AccessRequest, AccessResponse, test)
        Schema.to_file(file_name)
        data = get_file_data()
        self.assertEqual(1, len(data["paths"]))

        Schema.save_rpc("backend", "auth", AccessRequest, AccessResponse, test)
        Schema.to_file(file_name)
        data = get_file_data()
        self.assertEqual(1, len(data["paths"]))

        Schema.save_rpc("backend", "abc", AccessRequest, AccessResponse, test)
        Schema.to_file(file_name)
        data = get_file_data()
        self.assertEqual(2, len(data["paths"]))

        # add empty Schema
        Schema.clear()
        Schema.to_file(file_name)
        data = get_file_data()

        self.assertEqual(2, len(data["paths"]))
        self.assertGreaterEqual(len(data["components"]["schemas"]), 2)
        data = get_file_data()
        self.assertEqual(2, len(data["paths"]))
        self.assertGreaterEqual(len(data["components"]["schemas"]), 2)

    def test_clear(self):
        self.assertEqual(0, len(Schema.models))
        self.assertEqual(0, len(Schema.rpcs))
        Schema.save_rpc("backend", "auth", AccessRequest, AccessResponse, test)

        self.assertEqual(2, len(Schema.models))
        self.assertEqual(1, len(Schema.rpcs))

        Schema.clear()
        self.assertEqual(0, len(Schema.models))
        self.assertEqual(0, len(Schema.rpcs))

    def test_rpc_doc(self):
        with self.assertRaises(AssertionError):
            Schema.save_rpc("backend", "auth", AccessRequest, AccessResponse, test1)
        Schema.save_rpc("backend", "auth", AccessRequest, AccessResponse, test1, False)

    def test_partial_functions(self):
        func = partial(test, x=1)
        Schema.save_message("message_topic", AccessRequest, func)
        self.assertIn("/message_topic/test", Schema.render_messages())

    def test_save(self):
        Schema.save_rpc("backend", "abc", AccessRequest, AccessRequest, test)
        Schema.save_message("abc", AccessRequest, test)
        self.assertEqual(1, len(Schema.messages))
        self.assertEqual(1, len(Schema.rpcs))
        num_models = len(Schema.models)

        os.environ["TESTING"] = "False"
        Schema.save_rpc("backend", "abc", AccessRequest, AccessResponse, test)
        Schema.save_message("abc", AccessRequest, test)
        self.assertEqual(1, len(Schema.messages))
        self.assertEqual(1, len(Schema.rpcs))
        self.assertEqual(num_models, len(Schema.models))
        os.environ["TESTING"] = "True"


class LiveTest(RPCServerTestMixin, OpenModuleCoreTestMixin):
    rpc_channels = ["backend"]
    topics = ["backend"]

    def setUp(self):
        super().setUp()
        self.server = RPCServer(config=self.zmq_config(), context=self.zmq_context())
        self.server.run_as_thread()
        try:
            os.remove(file_name)
        except:
            pass
        Schema.clear()

    def tearDown(self) -> None:
        super().tearDown()
        try:
            os.remove(file_name)
        except:
            pass
        Schema.clear()

    def test_save(self):
        self.server.register_handler("backend", "test", AccessRequest, AccessResponse, test)
        self.core.messages.register_handler("backend", AccessRequest, test)
        self.wait_for_rpc_server(self.server)
        self.wait_for_dispatcher(self.core.messages)
        self.assertEqual(1, len(Schema.messages))
        self.assertEqual(1, len(Schema.rpcs))

        self.server.register_handler("backend", "abc", AccessRequest, AccessResponse, test3, register_schema=False)
        self.core.messages.register_handler("abc", AccessRequest, test3, register_schema=False)
        self.wait_for_rpc_server(self.server)
        self.wait_for_dispatcher(self.core.messages)
        self.assertEqual(1, len(Schema.messages))
        self.assertEqual(1, len(Schema.rpcs))




import random
import string

NAME = "THIS_IS_MY_NAME"
RESOURCE = "test-resource"
VERSION = "test-version"
random_suffix = "".join(random.choices(string.ascii_letters, k=10))
BROKER_SUB = f"inproc://test-sub-{random_suffix}"
BROKER_PUB = f"inproc://test-pub-{random_suffix}"



import logging
from collections import defaultdict
from typing import Union, Optional, Callable, DefaultDict, List, Dict, TypeVar, Type, Generic

import zmq
from pydantic import ValidationError, BaseModel

from openmodule.models.base import ZMQMessage
from openmodule.utils.schema import Schema


class Listener:
    def __init__(self, message_class: Type[ZMQMessage], filter: Optional[Dict], handler: Callable):
        self.filter = filter
        self.handler = handler
        self.message_class = message_class

    def matches(self, message: Dict):
        if self.filter is None:
            return True
        else:
            return message.items() >= self.filter.items()


EventArgs = TypeVar("EventArgs")


class EventListener(Generic[EventArgs], list):
    log: Optional[logging.Logger]

    def __init__(self, *args, log=None, raise_exceptions=False):
        super().__init__(args)
        self.raise_exceptions = raise_exceptions
        self.log = log or logging

    def __call__(self, *args: EventArgs):
        for f in self:
            try:
                f(*args)
            except zmq.ContextTerminated:
                raise
            except Exception as e:
                if self.raise_exceptions:
                    raise
                else:
                    self.log.exception(e)


ZMQMessageSub = TypeVar('ZMQMessageSub', bound=ZMQMessage)


class MessageDispatcher:
    def __init__(self, name=None, *, raise_validation_errors=False, raise_handler_errors=False):
        """
        :param name: optionally name the dispatcher for logging purposes
        :param raise_validation_errors: if true and received messages do not match a validation error is raised,
                                        this is useful in restricive code or testcases
        :param raise_handler_errors: if true and a message handler raises an exception, the exception is raised,
                                     this is useful in restricive code or testcases
        """
        self.name = name
        self.log = logging.getLogger(f"{self.__class__.__name__}({self.name})")
        self.listeners: DefaultDict[bytes, List[Listener]] = defaultdict(list)
        self.raise_validation_errors = raise_validation_errors
        self.raise_handler_errors = raise_handler_errors

    def unregister_handler(self, listener: Listener):
        for topic, listeners in self.listeners.items():
            try:
                listeners.remove(listener)
            except ValueError:
                pass

    def register_handler(self, topic: Union[bytes, str],
                         message_class: Type[ZMQMessageSub],
                         handler: Callable[[ZMQMessageSub], None], *,
                         filter: Optional[Dict] = None,
                         match_type=False,
                         register_schema=True):
        """
        registers a message handler. without any filters all messages from the topic are
        sent to the message handler.
        :param filter: a dictionary of values which must match in order for the message to
                       be further processed
        :param match_type: if set to true the message_class's type field is used as a filter.
                           equivalent to setting filter={"type": message_class.fields["type"].default}
        """

        if hasattr(topic, "encode"):
            topic = topic.encode()

        if match_type:
            assert "type" in message_class.__fields__ and message_class.__fields__["type"].default, (
                "Your message class definition does not set a `type` field, or the type field "
                "does not have a default value!"
            )
            filter = filter or {}
            filter["type"] = message_class.__fields__["type"].default

        listener = Listener(message_class, filter, handler)
        self.listeners[topic].append(listener)

        if register_schema:
            Schema.save_message(topic, message_class, handler, filter)

        return listener

    def dispatch(self, topic: bytes, message: Union[Dict, BaseModel]):
        if isinstance(message, BaseModel):
            message = message.dict()

        listeners = self.listeners.get(topic, [])
        for listener in listeners:
            if listener.matches(message):
                self.execute(listener, message)

    def execute(self, listener: Listener, message: Dict):
        try:
            parsed_message = listener.message_class(**message)
        except ValidationError as e:
            if self.raise_validation_errors:
                raise e from None
            else:
                self.log.exception("Invalid message received")
        else:
            try:
                listener.handler(parsed_message)
            except zmq.ContextTerminated:
                raise
            except Exception as e:
                if self.raise_handler_errors:
                    raise e from None
                else:
                    self.log.exception("Error in message handler")


class SubscribingMessageDispatcher(MessageDispatcher):
    def __init__(self, subscribe: Callable[[bytes], None], name=None, *, raise_validation_errors=False,
                 raise_handler_errors=False, unsubscribe: Optional[Callable[[bytes], None]] = None):
        super().__init__(name=name, raise_validation_errors=raise_validation_errors,
                         raise_handler_errors=raise_handler_errors)
        self.subscribe = subscribe
        self.unsubscribe = unsubscribe

    def register_handler(self, topic: Union[bytes, str],
                         message_class: Type[ZMQMessageSub],
                         handler: Callable[[ZMQMessageSub], None], *,
                         filter: Optional[Dict] = None,
                         register_schema=True,
                         match_type=False):
        if topic and hasattr(topic, "encode"):
            topic = topic.encode()
        self.subscribe(topic)
        return super().register_handler(topic, message_class, handler, filter=filter,
                                        register_schema=register_schema, match_type=match_type)

    def unregister_handler(self, listener: Listener):
        super().unregister_handler(listener)

        warn_no_unsubscribe = False
        for topic, listeners in self.listeners.items():
            if not listeners:
                if self.unsubscribe:
                    self.unsubscribe(topic)
                else:
                    warn_no_unsubscribe = True
                    break

        if warn_no_unsubscribe:
            self.log.warning("All handlers were unregistered from a topic, but no unsubscribe method was configured. "
                             "This may cause a performance overhead if too many unused subscriptions are kept. "
                             "Consider passing a unsubscribe method.")


class ZMQMessageDispatcher(SubscribingMessageDispatcher):
    def __init__(self, sub_socket: zmq.Socket, name=None, *, raise_validation_errors=False, raise_handler_errors=False):
        super().__init__(
            subscribe=lambda x: sub_socket.subscribe(x),
            unsubscribe=lambda x: sub_socket.unsubscribe(x),
            name=name,
            raise_validation_errors=raise_validation_errors,
            raise_handler_errors=raise_handler_errors
        )



import importlib.util
import inspect
import logging
import os
import random
import warnings
from contextlib import contextmanager
from socket import gethostname
from string import ascii_letters
from typing import Optional, Type, TypeVar

import functools
import builtins
import yaml
from pydantic import ValidationError
from pydantic.main import BaseModel


def _save_cast(obj, key, to_type, default):
    try:
        return to_type(obj[key])
    except (ValueError, TypeError, KeyError):
        return default


def int(key: str, default: int = 0) -> int:
    assert isinstance(default, (builtins.int,)), "default argument must be of type int"
    return _save_cast(os.environ, key, builtins.int, builtins.int(default))


def float(key: str, default: float = 0.0) -> float:
    assert isinstance(default, (builtins.int, builtins.float)), "default argument must be of type int or float"
    return _save_cast(os.environ, key, builtins.float, builtins.float(default))


def string(key: str, default: str = "") -> str:
    assert isinstance(default, str), "default argument must be of type string"
    res = os.environ.get(key, default)
    invalid_chars = "\"' \r\n\t"
    res = res.strip(invalid_chars)
    return res


def bool(key: str, default: bool = False) -> bool:
    assert isinstance(default, builtins.bool), "default argument must be of type bool"
    val = get(key) or ""
    val = val.upper().strip('\'" .,;-_')
    if not val:
        return default
    return val in ["T", "TRUE", "1", "Y", "YES", "J", "JA", "ON"]


def get(key: str, default: Optional[str] = "") -> str:
    return os.environ.get(key, default)


_inside_docker_cached = None


def _inside_docker():
    global _inside_docker_cached
    if _inside_docker_cached is not None:
        return _inside_docker_cached
    else:
        if os.path.exists('/.dockerenv'):
            _inside_docker_cached = True
            return _inside_docker_cached

        path = '/proc/self/cgroup'
        if os.path.isfile(path):
            with open(path, "r") as f:
                _inside_docker_cached = any('docker' in line for line in f)
        else:
            _inside_docker_cached = False

        return _inside_docker_cached


def debug():
    if "DEBUG" in os.environ:
        return bool("DEBUG", default=False)
    elif testing():
        return False
    else:
        return version() == "unknown" and not _inside_docker()


def settings_module():
    if testing() and os.path.exists("../tests/config.py"):
        return string("SETTINGS_MODULE", "tests.config")
    return string("SETTINGS_MODULE", "config")


def log_level() -> int:
    log_levels = {
        "INFO": logging.INFO,
        "DEBUG": logging.DEBUG,
        "VERBOSE": logging.DEBUG,
        "WARN": logging.WARNING,
        "WARNING": logging.WARNING,
        "ERROR": logging.ERROR,
        "ERR": logging.ERROR,
        "CRITICAL": logging.CRITICAL,
        "CRIT": logging.CRITICAL,
        "FATAL": logging.FATAL
    }
    default_log_level = logging.DEBUG if debug() else logging.INFO
    return log_levels.get(os.environ.get("LOG_LEVEL", "").upper(), default_log_level)


def version() -> str:
    version = string("VERSION", "")
    if not version:
        if os.path.isfile("VERSION"):
            with open("VERSION", "r") as f:
                version = f.read()
    if not version:
        version = "unknown"
    return version.strip("\r\n\t ")


def resource() -> str:
    resource = string("AUTH_RESOURCE", "")
    if not resource:
        # in test mode we allow using hostname, to make sure that the CI execution (docker) behaves like the local
        if testing() or not _inside_docker():
            resource = gethostname()
    return resource


def broker_pub(default: str = "tcp://127.0.0.1:10200") -> str:
    broker_host = string("BROKER_HOST", "")
    broker_pub_port = int("BROKER_PUB_PORT", 0)
    broker_pub = string("BROKER_PUB", "")

    if broker_pub:
        return broker_pub
    elif broker_host and broker_pub_port:
        # use legacy configs
        return "tcp://{}:{}".format(broker_host, broker_pub_port)
    else:
        return default


def broker_sub(default: str = "tcp://127.0.0.1:10100") -> str:
    broker_host = string("BROKER_HOST", "")
    broker_sub_port = int("BROKER_SUB_PORT", 0)
    broker_sub = string("BROKER_SUB", "")

    if broker_sub:
        return broker_sub
    elif broker_host and broker_sub_port:
        # use legacy configs
        return "tcp://{}:{}".format(broker_host, broker_sub_port)
    else:
        return default


def validate_config_module(config):
    required_keys = ["NAME", "RESOURCE", "VERSION", "DEBUG", "TESTING"]
    for x in required_keys:
        assert hasattr(config, x), f"the config module requires the variable {x}"

    if hasattr(config, "BROKER_PUB_PORT") or hasattr(config, "BROKER_SUB_PORT"):
        warnings.warn(
            "BROKER_(P/S)UB_PORT and BROKER_HOST are to be replaced with BROKER_(P/S)UB. In order to \n"
            "remain backwards compatible, you can use BROKER_PUB = config.broker_pub() which correctly \n"
            "interprets the deprecated environment variables",
            DeprecationWarning
        )


class SafeLoaderIgnoreUnknown(yaml.SafeLoader):
    def ignore_unknown(self, node):
        return None


SafeLoaderIgnoreUnknown.add_constructor(None, SafeLoaderIgnoreUnknown.ignore_unknown)

YamlType = TypeVar("YamlType", bound=BaseModel)


def config_yaml_path():
    yaml_path = string("CONFIG_YAML", "")
    if not yaml_path:
        if debug():
            yaml_path = "../settings/default-debug.yml"
        elif testing():
            yaml_path = "../settings/default-testing.yml"
        else:
            yaml_path = "/data/config.yml"
    return yaml_path


def yaml(model: Type[YamlType], path: str = None) -> YamlType:
    yaml_path = path or config_yaml_path()
    try:
        if os.path.exists(yaml_path):
            import yaml
            with open(yaml_path, "r") as f:
                a = yaml.load(f, SafeLoaderIgnoreUnknown)
                return model.parse_obj(a or {})
        else:
            return model()
    except ValidationError as e:
        logging.exception("error during config yaml loading, something is wrong with the configuration")
        raise e from None


def testing():
    return bool("TESTING", False)


def database_folder() -> str:
    path = string("DATABASE_FOLDER", f"/data/sqlite/")
    if testing():
        path = "../sqlite/test/"
    elif debug():
        path = "../sqlite/debug/"
    return path


def run_checks() -> bool:
    return debug() or testing()


empty = object()


def new_method_proxy(func):
    def inner(self, *args):
        if self._wrapped is empty:
            self._setup()
        return func(self._wrapped, *args)

    return inner


def unpickle_lazyobject(wrapped):
    return wrapped


class LazyObject:
    _wrapped = None

    def __init__(self):
        self._wrapped = empty

    __getattr__ = new_method_proxy(getattr)

    def __setattr__(self, name, value):
        if name == "_wrapped":
            # Assign to __dict__ to avoid infinite __setattr__ loops.
            self.__dict__["_wrapped"] = value
        else:
            if self._wrapped is empty:
                self._setup()
            setattr(self._wrapped, name, value)

    def __delattr__(self, name):
        assert name != "_wrapped", "can't delete _wrapped"
        if self._wrapped is empty:
            self._setup()
        delattr(self._wrapped, name)


class Settings:
    def __init__(self, lazy_setting, settings_module):
        self._explicit_settings = set()

        # update this dict from global settings (but only for ALL_CAPS settings)
        options = dict()
        for setting in dir(GlobalSettings):
            if not setting.startswith("_") and setting.isupper():
                options[setting] = getattr(GlobalSettings, setting)
                setattr(self, setting, options[setting])

        # store the settings module in case someone later cares
        self.SETTINGS_MODULE = settings_module
        if settings_module:
            for setting in dir(settings_module):
                if setting.isupper():
                    setting_value = getattr(settings_module, setting)
                    options[setting] = setting_value
                    setattr(self, setting, setting_value)

        override = _override_settings.set_base(lazy_setting, options)
        for k, v in override.items():
            setattr(self, k, v)
            self._explicit_settings.add(k)

    def is_overridden(self, setting):
        return setting in self._explicit_settings

    def __repr__(self):
        return '<%(cls)s "%(settings_module)s">' % {
            'cls': self.__class__.__name__,
            'settings_module': self.SETTINGS_MODULE,
        }


class LazySettings(LazyObject):
    _settings_module = None
    _is_configured = False

    def __init__(self):
        super().__init__()

    def _setup(self):
        module_name = settings_module()
        module = importlib.import_module(module_name)
        self._wrapped = Settings(self, module)

    def __repr__(self):
        # Hardcode the class name as otherwise it yields 'Settings'.
        if self._wrapped is empty:
            return '<LazySettings [Unevaluated]>'
        return '<LazySettings "%(settings_module)s">' % {
            'settings_module': self._wrapped.SETTINGS_MODULE,
        }

    def __getattr__(self, name):
        """Return the value of a setting and cache it in self.__dict__."""
        if name == "_is_configured":
            return self._is_configured
        if self._wrapped is empty:
            self._setup()
        val = getattr(self._wrapped, name)
        self.__dict__[name] = val
        return val

    def __setattr__(self, name, value):
        if name == '_wrapped':
            self.__dict__.clear()
        else:
            self.__dict__.pop(name, None)
        super().__setattr__(name, value)

    def __delattr__(self, name):
        """Delete a setting and clear it from cache if needed."""
        super().__delattr__(name)
        self.__dict__.pop(name, None)

    def configure(self, settings_module):
        assert self._wrapped is empty, 'Settings are already configured'
        self._wrapped = Settings(self, settings_module)

    def reset(self):
        self._wrapped = empty
        _override_settings.clear_all()

    def override(self, **options):
        _override_settings.override(**options)

    def clear_override(self, **options):
        _override_settings.clear_override(**options)


class OverrideSettings:
    lazy_setting = None

    def __init__(self):
        self.overrides = dict()

    def clear_all(self):
        self.lazy_setting = None
        self.overrides = dict()

    def override(self, **options):
        for k, v in options.items():
            if self.overrides.get(k):
                self.overrides[k].append(v)
            else:
                self.overrides[k] = [v]
            if self.lazy_setting is not None:
                setattr(self.lazy_setting, k, v)
                self.lazy_setting._wrapped._explicit_settings.add(k)

    def clear_override(self, **options):
        for k in options.keys():
            if self.overrides.get(k):
                self.overrides[k].pop()
                if self.lazy_setting is not None:
                    if len(self.overrides[k]) == 1:
                        self.lazy_setting._wrapped._explicit_settings.remove(k)
                    setattr(self.lazy_setting, k, self.overrides[k][-1])

    def set_base(self, lazy_setting, options):
        assert self.lazy_setting is None, "Do not use multiple settings"
        self.lazy_setting = lazy_setting
        overrides = {}
        for k, v in options.items():
            if self.overrides.get(k):
                overrides[k] = self.overrides[k][-1]
                self.overrides[k].insert(0, v)
            else:
                self.overrides[k] = [v]
        return overrides


@contextmanager
def override_context(**options):
    settings.override(**options)
    yield
    settings.clear_override(**options)


def override_settings(**options):
    def decorator(x):
        if inspect.isclass(x):
            original_run = x.run

            def new_run(inner_self, result=None):
                with override_context(**options):
                    original_run(inner_self, result)

            x.run = new_run
            return x
        elif callable(x):
            @functools.wraps(x)
            def wrapper(*args, **kwargs):
                with override_context(**options):
                    x(*args, **kwargs)

            return wrapper

    return decorator


class GlobalSettings:
    # usual
    NAME = string("NAME", "om_dev_unnamed_1")
    VERSION = version()
    RESOURCE = resource()
    DEBUG = debug()
    TESTING = testing()
    LOG_LEVEL = log_level()
    DATABASE_FOLDER = database_folder()

    # broker env vars
    BROKER_SUB = broker_sub()
    BROKER_PUB = broker_pub()

    LOCAL_DEVELOPMENT = bool("LOCAL_DEVELOPMENT", False)


settings = LazySettings()
_override_settings = OverrideSettings()



from typing import Dict

from openmodule.config import settings
from openmodule.core import init_openmodule, shutdown_openmodule, OpenModuleCore
from openmodule_test.health import HealthTestMixin


class OpenModuleCoreTestMixin(HealthTestMixin):
    """
    Mixin which creates a core, zmq, and health mixin
    """

    init_kwargs: Dict = {}
    core: OpenModuleCore

    def get_init_kwargs(self):
        return self.init_kwargs

    def setUp(self):
        super().setUp()
        self.init_kwargs.setdefault("sentry", False)
        self.init_kwargs.setdefault("dsgvo", False)
        self.core = init_openmodule(
            config=self.zmq_config(),
            context=self.zmq_context(),
            **self.get_init_kwargs()
        )
        self.addCleanup(shutdown_openmodule)
        self.zmq_client.subscribe(b"healthz")
        self.wait_for_health(self.core.config.NAME)



import random
import string
import threading
import time
from contextlib import suppress
from functools import partial

from pydantic.main import BaseModel

from openmodule_test.zeromq import ZMQTestMixin


class _EmptyModel(BaseModel):
    pass


class RPCServerTestMixin(ZMQTestMixin):
    def wait_for_rpc_server(self, server):
        message_received = False

        def handler(_, __):
            nonlocal message_received
            message_received = True

        """
        waits until a rpc server is responding on the last channel we registered
        this assumes that the subscription we issue is the last and if it is connected, 
        all previous subscriptions will also be connected
        """
        assert server.handlers, "you need to register the handlers beforehand"
        random_channel = "_test" + "".join(random.choices(string.ascii_letters, k=10))

        server.register_handler(random_channel, "ping", _EmptyModel, _EmptyModel, handler, register_schema=False)

        for x in range(self.zmq_client.startup_check_iterations):
            self.rpc(random_channel, "ping", {}, receive_response=False)
            time.sleep(self.zmq_client.startup_check_delay)
            if message_received:
                break

        assert message_received, "error during startup and connect"
